# -*- coding: utf-8 -*-
#
#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from array import array
from CachedMethods import cached_args_method
from collections import Counter, defaultdict
from functools import cached_property
from itertools import zip_longest
from math import ceil
from numpy import uint, zeros
from struct import pack_into, unpack_from
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union
from weakref import ref
from zlib import compress, decompress
from . import cgr, query  # cyclic imports resolve
from .bonds import Bond, DynamicBond, QueryBond
from .graph import Graph
from ..algorithms.aromatics import Aromatize
from ..algorithms.calculate2d import Calculate2DMolecule
from ..algorithms.depict import DepictMolecule
from ..algorithms.fingerprints import Fingerprints
from ..algorithms.huckel import Huckel
from ..algorithms.mcs import MCS
from ..algorithms.smiles import MoleculeSmiles
from ..algorithms.standardize import Saturation, StandardizeMolecule
from ..algorithms.stereo import MoleculeStereo
from ..algorithms.tautomers import Tautomers
from ..algorithms.x3dom import X3domMolecule
from ..exceptions import MappingError, ValenceError
from ..periodictable import DynamicElement, Element, QueryElement


class MoleculeContainer(MoleculeStereo, Graph[Element, Bond], Aromatize, StandardizeMolecule, MoleculeSmiles,
                        DepictMolecule, Calculate2DMolecule, Fingerprints, Tautomers, MCS, Huckel,
                        Saturation, X3domMolecule):
    __slots__ = ('_conformers', '_atoms_stereo', '_hydrogens', '_cis_trans_stereo', '_allenes_stereo',
                 '_parsed_mapping', '_backup', '__meta', '__name')

    def __init__(self):
        super().__init__()
        self._conformers: List[Dict[int, Tuple[float, float, float]]] = []
        self._hydrogens: Dict[int, Optional[int]] = {}
        self._atoms_stereo: Dict[int, bool] = {}
        self._allenes_stereo: Dict[int, bool] = {}
        self._cis_trans_stereo: Dict[Tuple[int, int], bool] = {}
        self._parsed_mapping: Dict[int, int] = {}
        self.__meta = None
        self.__name = None

    @property
    def meta(self) -> Dict:
        if self.__meta is None:
            self.__meta = {}  # lazy
        return self.__meta

    @property
    def name(self) -> str:
        return self.__name or ''

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError('name should be string up to 80 symbols')
        self.__name = name

    @cached_args_method
    def environment(self, atom: int, include_bond: bool = True, include_atom: bool = True) -> \
            Tuple[Union[Tuple[int, Bond, Element],
                        Tuple[int, Element],
                        Tuple[int, Bond],
                        int], ...]:
        """
        groups of (atom_number, bond, atom) connected to atom or
        groups of (atom_number, bond) connected to atom or
        groups of (atom_number, atom) connected to atom or
        neighbors atoms connected to atom

        :param atom: number
        :param include_atom: include atom object
        :param include_bond: include bond object
        """
        if include_atom:
            atoms = self._atoms
            if include_bond:
                return tuple((n, bond, atoms[n]) for n, bond in self._bonds[atom].items())
            return tuple((n, atoms[n]) for n in self._bonds[atom])
        elif include_bond:
            return tuple(self._bonds[atom].items())
        return tuple(self._bonds[atom])

    @cached_args_method
    def neighbors(self, n: int) -> int:
        """number of neighbors atoms excluding any-bonded"""
        return sum(b.order != 8 for b in self._bonds[n].values())

    @cached_args_method
    def hybridization(self, n: int) -> int:
        """
        Atom hybridization.

        1 - if atom has zero or only single bonded neighbors, 2 - if has only one double bonded neighbor and any amount
        of single bonded, 3 - if has one triple bonded and any amount of double and single bonded neighbors or
        two and more double bonded and any amount of single bonded neighbors, 4 - if atom in aromatic ring.
        """
        hybridization = 1
        for bond in self._bonds[n].values():
            order = bond.order
            if order == 4:
                return 4
            elif order == 3:
                if hybridization != 3:
                    hybridization = 3
            elif order == 2:
                if hybridization == 1:
                    hybridization = 2
                elif hybridization == 2:
                    hybridization = 3
        return hybridization

    @cached_args_method
    def heteroatoms(self, n: int) -> int:
        """
        Number of neighbored heteroatoms (not carbon or hydrogen)
        """
        atoms = self._atoms
        return sum(atoms[m].atomic_number not in (1, 6) for m in self._bonds[n])

    def implicit_hydrogens(self, n: int) -> Optional[int]:
        """
        Number of implicit hydrogen atoms connected to atom.

        Returns None if count are ambiguous.
        """
        return self._hydrogens[n]

    @cached_args_method
    def explicit_hydrogens(self, n: int) -> int:
        """
        Number of explicit hydrogen atoms connected to atom.

        Take into account any type of bonds with hydrogen atoms.
        """
        atoms = self._atoms
        return sum(atoms[m].atomic_number == 1 for m in self._bonds[n])

    @cached_args_method
    def total_hydrogens(self, n: int) -> int:
        """
        Number of hydrogen atoms connected to atom.

        Take into account any type of bonds with hydrogen atoms.
        """
        return self._hydrogens[n] + self.explicit_hydrogens(n)

    def adjacency_matrix(self, set_bonds=False):
        """
        Adjacency matrix of Graph.

        :param set_bonds: if True set bond orders instead of 1.
        """
        adj = zeros((len(self), len(self)), dtype=uint)
        mapping = {n: x for x, n in enumerate(self._atoms)}
        if set_bonds:
            for n, ms in self._bonds.items():
                n = mapping[n]
                for m, b in ms.items():
                    adj[n, mapping[m]] = int(b)
        else:
            for n, ms in self._bonds.items():
                n = mapping[n]
                for m, b in ms.items():
                    adj[n, mapping[m]] = 1
        return adj

    @cached_property
    def molecular_charge(self) -> int:
        """
        Total charge of molecule
        """
        return sum(self._charges.values())

    @cached_property
    def is_radical(self) -> bool:
        """
        True if at least one atom is radical
        """
        return any(self._radicals.values())

    @cached_property
    def molecular_mass(self) -> float:
        return sum(x.atomic_mass for x in self._atoms.values())

    @cached_property
    def brutto(self) -> Dict[str, int]:
        """Counted atoms dict"""
        return Counter(x.atomic_symbol for x in self._atoms.values())

    def add_atom(self, atom: Union[Element, int, str], *args, charge=0, is_radical=False, **kwargs):
        """
        Add new atom.
        """
        if not isinstance(atom, Element):
            if isinstance(atom, str):
                atom = Element.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = Element.from_atomic_number(atom)()
            else:
                raise TypeError('Element object expected')

        _map = super().add_atom(atom, *args, charge=charge, is_radical=is_radical, **kwargs)
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        if atom.atomic_number != 1:
            try:
                rules = atom.valence_rules(charge, is_radical, 0)
            except ValenceError:
                self._hydrogens[_map] = None
            else:
                self._hydrogens[_map] = rules[0][2]  # first rule without neighbors
        else:
            self._hydrogens[_map] = 0
        return _map

    def add_bond(self, n, m, bond: Union[Bond, int]):
        """
        Connect atoms with bonds.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.
        """
        if not isinstance(bond, Bond):
            bond = Bond(bond)

        super().add_bond(n, m, bond)
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        self._calc_implicit(n)
        self._calc_implicit(m)

        if self._atoms[n].atomic_number != 1 and self._atoms[m].atomic_number != 1:  # not hydrogen
            # fix stereo if formed not to hydrogen bond
            self.fix_stereo()

    def delete_atom(self, n):
        """
        Remove atom.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.
        """
        old_bonds = self._bonds[n]  # save bonds
        isnt_hydrogen = self._atoms[n].atomic_number != 1
        super().delete_atom(n)

        del self._hydrogens[n]
        self._conformers.clear()  # clean conformers. need full recalculation for new system
        try:
            del self._parsed_mapping[n]
        except KeyError:
            pass

        for m in old_bonds:
            self._calc_implicit(m)

        if isnt_hydrogen:  # hydrogen atom not used for stereo coding
            self.fix_stereo()

    def delete_bond(self, n, m):
        """
        Disconnect atoms.

        For Thiele forms of molecule causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.
        """
        super().delete_bond(n, m)
        self._conformers.clear()  # clean conformers. need full recalculation for new system

        self._calc_implicit(n)
        self._calc_implicit(m)

        if self._atoms[n].atomic_number != 1 and self._atoms[m].atomic_number != 1:
            self.fix_stereo()

    def remap(self, mapping: Dict[int, int], *, copy: bool = False) -> 'MoleculeContainer':
        if len(mapping) != len(set(mapping.values())) or \
                not (self._atoms.keys() - mapping.keys()).isdisjoint(mapping.values()):
            raise ValueError('mapping overlap')

        mg = mapping.get
        sp = self._plane
        sc = self._charges
        sr = self._radicals
        shg = self._hydrogens

        if copy:
            h = self.__class__()
            h._MoleculeContainer__name = self.__name
            if self.__meta is not None:
                h._MoleculeContainer__meta = self.__meta.copy()
            hb = h._bonds
            ha = h._atoms
            hc = h._charges
            hr = h._radicals
            hp = h._plane
            hhg = h._hydrogens
            hcf = h._conformers
            has = h._atoms_stereo
            hal = h._allenes_stereo
            hcs = h._cis_trans_stereo
            hm = h._parsed_mapping

            for n, atom in self._atoms.items():
                m = mg(n, n)
                atom = atom.copy()
                ha[m] = atom
                atom._attach_to_graph(h, m)

            # deep copy of bonds
            for n, m_bond in self._bonds.items():
                n = mg(n, n)
                hb[n] = hbn = {}
                for m, bond in m_bond.items():
                    m = mg(m, m)
                    if m in hb:  # bond partially exists. need back-connection.
                        hbn[m] = hb[m][n]
                    else:
                        hbn[m] = bond.copy()
        else:
            hb = {}
            ha = {}
            hc = {}
            hr = {}
            hp = {}
            hhg = {}
            hcf = []
            has = {}
            hal = {}
            hcs = {}
            hm = {}
            for n, atom in self._atoms.items():
                m = mg(n, n)
                ha[m] = atom
                atom._change_map(m)  # change mapping number

            for n, m_bond in self._bonds.items():
                hb[mg(n, n)] = {mg(m, m): b for m, b in m_bond.items()}

        for n in self._atoms:
            m = mg(n, n)
            hc[m] = sc[n]
            hr[m] = sr[n]
            hp[m] = sp[n]
            hhg[m] = shg[n]

        hcf.extend({mg(n, n): x for n, x in c.items()} for c in self._conformers)
        for n, m in self._parsed_mapping.items():
            hm[mg(n, n)] = m
        for n, stereo in self._atoms_stereo.items():
            has[mg(n, n)] = stereo
        for n, stereo in self._allenes_stereo.items():
            hal[mg(n, n)] = stereo
        for (n, m), stereo in self._cis_trans_stereo.items():
            hcs[(mg(n, n), mg(m, m))] = stereo

        if copy:
            return h

        self._atoms = ha
        self._bonds = hb
        self._charges = hc
        self._radicals = hr
        self._plane = hp
        self._hydrogens = hhg
        self._conformers = hcf
        self._atoms_stereo = has
        self._allenes_stereo = hal
        self._cis_trans_stereo = hcs
        self._parsed_mapping = hm
        self.flush_cache()
        return self

    def copy(self) -> 'MoleculeContainer':
        copy = super().copy()
        copy._MoleculeContainer__name = self.__name
        if self.__meta is None:
            copy._MoleculeContainer__meta = None
        else:
            copy._MoleculeContainer__meta = self.__meta.copy()
        copy._hydrogens = self._hydrogens.copy()
        copy._parsed_mapping = self._parsed_mapping.copy()
        copy._conformers = [c.copy() for c in self._conformers]
        copy._atoms_stereo = self._atoms_stereo.copy()
        copy._allenes_stereo = self._allenes_stereo.copy()
        copy._cis_trans_stereo = self._cis_trans_stereo.copy()
        return copy

    def union(self, other: 'MoleculeContainer', *, remap=False) -> 'MoleculeContainer':
        """
        :param remap: if atoms has collisions then remap other graph atoms else raise exception.
        """
        if not isinstance(other, MoleculeContainer):
            raise TypeError('MoleculeContainer expected')
        elif self._atoms.keys() & other._atoms.keys():
            if remap:
                other = other.remap({n: i for i, n in enumerate(other, start=max(self._atoms) + 1)}, copy=True)
            else:
                raise MappingError('mapping of graphs is not disjoint')
        u = super().union(other)
        u._MoleculeContainer__name = u._MoleculeContainer__meta = None
        u._conformers.clear()
        u._hydrogens.update(other._hydrogens)
        u._parsed_mapping.update(other._parsed_mapping)
        u._atoms_stereo.update(other._atoms_stereo)
        u._allenes_stereo.update(other._allenes_stereo)
        u._cis_trans_stereo.update(other._cis_trans_stereo)
        return u

    def substructure(self, atoms: Iterable[int], *, as_query: bool = False, recalculate_hydrogens=True,
                     skip_neighbors_marks=False, skip_hybridizations_marks=False, skip_hydrogens_marks=False,
                     skip_rings_sizes_marks=False,) -> Union['MoleculeContainer', 'query.QueryContainer']:
        """
        Create substructure containing atoms from atoms list.

        For Thiele forms of molecule In Molecule substructure causes invalidation of internal state.
        Implicit hydrogens marks will not be set if atoms in aromatic rings.
        Call `kekule()` and `thiele()` in sequence to fix marks.

        :param atoms: list of atoms numbers of substructure
        :param as_query: return Query object based on graph substructure
        :param recalculate_hydrogens: calculate implicit H count in substructure
        :param skip_neighbors_marks: Don't set neighbors count marks on substructured queries
        :param skip_hybridizations_marks: Don't set hybridizations marks on substructured queries
        :param skip_hydrogens_marks: Don't set hydrogens count marks on substructured queries
        :param skip_rings_sizes_marks: Don't set rings_sizes marks on substructured queries
        """
        if not atoms:
            raise ValueError('empty atoms list not allowed')
        if set(atoms) - self._atoms.keys():
            raise ValueError('invalid atom numbers')
        atoms = tuple(n for n in self._atoms if n in atoms)  # save original order
        if as_query:
            atom_type = QueryElement
            bond_type = QueryBond
            sub = object.__new__(query.QueryContainer)
        else:
            atom_type = Element
            bond_type = Bond
            sub = object.__new__(self.__class__)
            sub._MoleculeContainer__name = sub._MoleculeContainer__meta = None

        sa = self._atoms
        sb = self._bonds
        sc = self._charges
        sr = self._radicals
        sp = self._plane

        sub._charges = {n: sc[n] for n in atoms}
        sub._radicals = {n: sr[n] for n in atoms}
        sub._plane = {n: sp[n] for n in atoms}

        sub._atoms = ca = {}
        for n in atoms:
            ca[n] = atom = atom_type.from_atom(sa[n])
            atom._attach_to_graph(sub, n)

        sub._bonds = cb = {}
        for n in atoms:
            cb[n] = cbn = {}
            for m, bond in sb[n].items():
                if m in cb:  # bond partially exists. need back-connection.
                    cbn[m] = cb[m][n]
                elif m in atoms:
                    cbn[m] = bond_type.from_bond(bond)

        if as_query:
            lost = {n for n, a in sa.items() if a.atomic_number != 1} - set(atoms)  # atoms not in substructure
            not_skin = {n for n in atoms if lost.isdisjoint(sb[n])}
            sub._atoms_stereo = {n: s for n, s in self._atoms_stereo.items() if n in not_skin}
            sub._allenes_stereo = {n: s for n, s in self._allenes_stereo.items()
                                   if not_skin.issuperset(self._stereo_allenes_paths[n]) and
                                      not_skin.issuperset(x for x in self._stereo_allenes[n] if x)}
            sub._cis_trans_stereo = {nm: s for nm, s in self._cis_trans_stereo.items()
                                     if not_skin.issuperset(self._stereo_cis_trans_paths[nm]) and
                                        not_skin.issuperset(x for x in self._stereo_cis_trans[nm] if x)}

            sub._heteroatoms = {n: () for n in atoms}

            if skip_hybridizations_marks:
                sub._hybridizations = {n: () for n in atoms}
            else:
                sh = self.hybridization
                sub._hybridizations = {n: (sh(n),) for n in atoms}
            if skip_neighbors_marks:
                sub._neighbors = {n: () for n in atoms}
            else:
                sn = self.neighbors
                sub._neighbors = {n: (sn(n),) for n in atoms}
            if skip_hydrogens_marks:
                sub._hydrogens = {n: () for n in atoms}
            else:
                shg = self._hydrogens
                sub._hydrogens = {n: () if shg[n] is None else (shg[n],) for n in atoms}
            if skip_rings_sizes_marks:
                sub._rings_sizes = {n: () for n in atoms}
            else:
                rs = self.atoms_rings_sizes
                sub._rings_sizes = {n: rs.get(n, ()) for n in atoms}
        else:
            sub._conformers = [{n: c[n] for n in atoms} for c in self._conformers]

            if recalculate_hydrogens:
                sub._hydrogens = {}
                for n in atoms:
                    sub._calc_implicit(n)
            else:
                hg = self._hydrogens
                sub._hydrogens = {n: hg[n] for n in atoms}

            sub._parsed_mapping = {n: m for n, m in self._parsed_mapping.items() if n in atoms}

            # fix_stereo will repair data
            sub._atoms_stereo = self._atoms_stereo.copy()
            sub._allenes_stereo = self._allenes_stereo.copy()
            sub._cis_trans_stereo = self._cis_trans_stereo.copy()
            sub.fix_stereo()
        return sub

    def augmented_substructure(self, atoms: Iterable[int], deep: int = 1, **kwargs) -> 'MoleculeContainer':
        """
        Create substructure containing atoms and their neighbors

        :param atoms: list of core atoms in graph
        :param deep: number of bonds between atoms and neighbors
        """
        return self.substructure(self._augmented_substructure(atoms, deep)[-1], **kwargs)

    def augmented_substructures(self, atoms: Iterable[int], deep: int = 1, **kwargs) -> List['MoleculeContainer']:
        """
        Create list of substructures containing atoms and their neighbors

        :param atoms: list of core atoms in graph
        :param deep: number of bonds between atoms and neighbors
        :return: list of graphs containing atoms, atoms + first circle, atoms + 1st + 2nd,
            etc up to deep or while new nodes available
        """
        return [self.substructure(a, **kwargs) for a in self._augmented_substructure(atoms, deep)]

    def split(self) -> List['MoleculeContainer']:
        """
        Split disconnected structure to connected substructures
        """
        return [self.substructure(c, recalculate_hydrogens=False) for c in self.connected_components]

    def compose(self, other: 'MoleculeContainer') -> 'cgr.CGRContainer':
        """
        Compose 2 graphs to CGR.
        """
        if not isinstance(other, MoleculeContainer):
            raise TypeError('MoleculeContainer expected')
        sa = self._atoms
        sc = self._charges
        sr = self._radicals
        sp = self._plane
        sb = self._bonds

        bonds = []
        adj = defaultdict(lambda: defaultdict(lambda: [None, None]))

        oa = other._atoms
        oc = other._charges
        or_ = other._radicals
        op = other._plane
        ob = other._bonds

        common = sa.keys() & oa.keys()

        h = cgr.CGRContainer()
        ha = h._atoms
        hb = h._bonds
        hc = h._charges
        hpc = h._p_charges
        hr = h._radicals
        hpr = h._p_radicals
        hp = h._plane

        for n in sa.keys() - common:  # cleavage atoms
            hc[n] = hpc[n] = sc[n]
            hr[n] = hpr[n] = sr[n]
            hp[n] = sp[n]
            hb[n] = {}
            ha[n] = a = DynamicElement.from_atom(sa[n])
            a._attach_to_graph(h, n)

            for m, bond in sb[n].items():
                if m not in ha:
                    if m in common:  # bond to common atoms is broken bond
                        bond = DynamicBond(bond.order, None)
                    else:
                        bond = DynamicBond(bond.order, bond.order)
                    bonds.append((n, m, bond))
        for n in oa.keys() - common:  # coupling atoms
            hc[n] = hpc[n] = oc[n]
            hr[n] = hpr[n] = or_[n]
            hp[n] = op[n]
            hb[n] = {}
            ha[n] = a = DynamicElement.from_atom(oa[n])
            a._attach_to_graph(h, n)

            for m, bond in ob[n].items():
                if m not in ha:
                    if m in common:  # bond to common atoms is formed bond
                        bond = DynamicBond(None, bond.order)
                    else:
                        bond = DynamicBond(bond.order, bond.order)
                    bonds.append((n, m, bond))
        for n in common:
            an = adj[n]
            for m, bond in sb[n].items():
                if m in common:
                    an[m][0] = bond.order
            for m, bond in ob[n].items():
                if m in common:
                    an[m][1] = bond.order
        for n in common:
            san = sa[n]
            if san.atomic_number != oa[n].atomic_number or san.isotope != oa[n].isotope:
                raise MappingError(f'atoms with number {{{n}}} not equal')

            hc[n] = sc[n]
            hpc[n] = oc[n]
            hr[n] = sr[n]
            hpr[n] = or_[n]
            hp[n] = sp[n]
            hb[n] = {}
            ha[n] = a = DynamicElement.from_atom(san)
            a._attach_to_graph(h, n)

            for m, (o1, o2) in adj[n].items():
                if m not in ha:
                    bonds.append((n, m, DynamicBond(o1, o2)))

        for n, m, bond in bonds:
            hb[n][m] = hb[m][n] = bond
        return h

    def get_fast_mapping(self, other: 'MoleculeContainer') -> Optional[Dict[int, int]]:
        """
        Get self to other fast (suboptimal) structure mapping.
        Only one possible atoms mapping returned.
        Effective only for big molecules.
        """
        if isinstance(other, MoleculeContainer):
            if len(self) != len(other):
                return
            so = self.smiles_atoms_order
            oo = other.smiles_atoms_order
            if self != other:
                return
            return dict(zip(so, oo))
        raise TypeError('MoleculeContainer expected')

    def get_mapping(self, other: 'MoleculeContainer', /, **kwargs):
        if isinstance(other, MoleculeContainer):
            return super().get_mapping(other, **kwargs)
        raise TypeError('MoleculeContainer expected')

    def pack(self) -> bytes:
        """
        Pack into compressed bytes.
        Note:
            * Less than 4096 atoms supported. Atoms mapping should be in range 1-4095.
            * Implicit hydrogens count should be in range 0-6 or unspecified.
            * Isotope shift should be in range -15 - 15 relatively mdl.common_isotopes
            * Atoms neighbors should be in range 0-15

        Format specification:
        Big endian bytes order
        8 bit - empty byte for future extending
        12 bit - number of atoms
        12 bit - cis/trans stereo block size
        Atom block 9 bytes (repeated):
        12 bit - atom number
        4 bit - number of neighbors
        2 bit tetrahedron sign (00 - not stereo, 10 or 11 - has stereo)
        2 bit - allene sign
        5 bit - isotope (00000 - not specified, over = isotope - common_isotope + 16)
        7 bit - atomic number (<=118)
        32 bit - XY float16 coordinates
        3 bit - hydrogens (0-7). Note: 7 == None
        4 bit - charge (charge + 4. possible range -4 - 4)
        1 bit - radical state
        Connection table: flatten list of neighbors. neighbors count stored in atom block.
        For example CC(=O)O - {1: [2], 2: [1, 3, 4], 3: [2], 4: [2]} >> [2, 1, 3, 4, 2, 2].
        Repeated block (equal to bonds count).
        24 bit - paired 12 bit numbers.
        Bonds order block (repeated):
        16 bit - 5 bonds grouped (3 bit each). 1 bit unused. Zero padding used than bonds count not proportional to 5.
        Cis/trans data block (repeated):
        24 bit - atoms pair
        7 bit - zero padding. in future can be used for extra bond-level stereo, like atropoisomers.
        1 bit - sign
        """
        bonds = self._bonds
        if max(bonds) > 4095:
            raise ValueError('Big molecules not supported')
        if any(len(x) > 15 for x in bonds.values()):
            raise ValueError('To many neighbors not supported')
        from ..files._mdl.mol import common_isotopes

        plane = self._plane
        charges = self._charges
        radicals = self._radicals
        hydrogens = self._hydrogens
        atoms_stereo = self._atoms_stereo
        allenes_stereo = self._allenes_stereo
        cis_trans_stereo = self._cis_trans_stereo

        data = bytearray(4 +  # extension byte + atoms count + cis/trans bit
                         9 * self.atoms_count +  # atoms data
                         3 * self.bonds_count +  # connection table
                         2 * ceil(self.bonds_count / 5) +  # bonds order
                         4 * len(cis_trans_stereo))
        pack_into('>HB', data, 1, (self.atoms_count << 4) | (len(cis_trans_stereo) >> 8), len(cis_trans_stereo) & 0xff)
        shift = 4

        neighbors = []
        bonds_pack = []
        seen = set()
        hold = []
        for o, (n, a) in enumerate(self._atoms.items()):
            bs = bonds[n]
            neighbors.extend(bs)
            seen.add(n)
            for m, b in bs.items():
                if m not in seen:
                    bonds_pack.append(b.order - 1)  # 8 - 4 bit, but 7 - 3 bit

            # 3 bit - hydrogens (0-7) | 4 bit - charge | 1 bit - radical
            hcr = (charges[n] + 4) << 1
            if radicals[n]:
                hcr |= 1
            if (h := hydrogens[n]) is None:
                hcr |= 224
            else:
                hcr |= h << 5

            # 2 bit tetrahedron sign | 2 bit - allene sign | 5 bit - isotope | 7 bit - atomic number (<=118)
            sia = a.atomic_number
            if a.isotope:
                sia |= (a.isotope - common_isotopes[a.atomic_symbol] + 16) << 7

            if n in atoms_stereo:
                if atoms_stereo[n]:
                    sia |= 0xc000
                else:
                    sia |= 0x8000
            if n in allenes_stereo:
                if allenes_stereo[n]:
                    sia |= 0x3000
                else:
                    sia |= 0x2000

            hold.append((shift + 9 * o, (n << 4) | len(bs), sia, *plane[n], hcr))

        shift += 9 * self.atoms_count + 3 * self.bonds_count - 4
        ngb = iter(reversed(neighbors))
        for o, (n2, n1) in enumerate(zip_longest(ngb, ngb)):
            # 12 bit + 12 bit
            pack_into('>I', data, shift - 3 * o, (n1 << 12) | n2)

        # pack after connection table for preventing override!
        for x in hold:
            pack_into('>2H2eB', data, *x)

        # 16 bit - 5 bonds packing. 1 bit empty.
        shift += 4
        bp = iter(bonds_pack)
        for o, (b1, b2, b3, b4, b5) in enumerate(zip_longest(bp, bp, bp, bp, bp, fillvalue=0)):
            pack_into('>H', data, shift + 2 * o, (b1 << 12) | (b2 << 9) | (b3 << 6) | (b4 << 3) | b5)

        shift += 2 * ceil(self.bonds_count / 5)
        for o, ((n, m), s) in enumerate(cis_trans_stereo.items()):
            pack_into('>I', data, shift + 4 * o, (n << 20) | (m << 8) | s)

        return compress(bytes(data), 9)

    @classmethod
    def unpack(cls, data: bytes) -> 'MoleculeContainer':
        """
        Unpack from compressed bytes.
        """
        try:  # windows? ;)
            from ._unpack import unpack
        except ImportError:
            return cls.pure_unpack(data)
        (mapping, atom_numbers, isotopes, charges, radicals, hydrogens, plane, bonds,
         atoms_stereo, allenes_stereo, cis_trans_stereo) = unpack(decompress(data))

        mol = object.__new__(cls)
        mol._bonds = bonds
        mol._plane = plane
        mol._charges = charges
        mol._radicals = radicals
        mol._hydrogens = hydrogens
        mol._atoms_stereo = atoms_stereo
        mol._allenes_stereo = allenes_stereo
        mol._cis_trans_stereo = cis_trans_stereo

        mol._conformers = []
        mol._parsed_mapping = {}
        mol._MoleculeContainer__meta = None
        mol._MoleculeContainer__name = None
        mol._atoms = atoms = {}

        for n, a, i in zip(mapping, atom_numbers, isotopes):
            atoms[n] = a = object.__new__(Element.from_atomic_number(a))
            a._Core__isotope = i
            a._graph = ref(mol)
            a._map = n
        return mol

    @classmethod
    def pure_unpack(cls, data: bytes) -> 'MoleculeContainer':
        """
        Unpack from compressed bytes. Python implementation.
        """
        from ..files._mdl.mol import common_isotopes

        data = memoryview(decompress(data))
        mol = cls()
        atoms = mol._atoms
        bonds = mol._bonds
        plane = mol._plane
        charges = mol._charges
        radicals = mol._radicals
        hydrogens = mol._hydrogens
        atoms_stereo = mol._atoms_stereo
        allenes_stereo = mol._allenes_stereo
        cis_trans_stereo = mol._cis_trans_stereo

        neighbors = {}
        acs = int.from_bytes(data[1:4], 'big')
        shift = 4
        for o in range(acs >> 12):
            nn, sia, x, y, hcr = unpack_from('>2H2eB', data, shift + 9 * o)
            n = nn >> 4
            neighbors[n] = nn & 0x0f
            # stereo
            s = sia >> 14
            if s:
                atoms_stereo[n] = s == 3
            s = (sia >> 12) & 3
            if s:
                allenes_stereo[n] = s == 3

            # atoms
            a = Element.from_atomic_number(sia & 0x7f)
            ai = (sia >> 7) & 0x1f
            if ai:
                ai += common_isotopes[a.__name__] - 16
            else:
                ai = None
            atoms[n] = a = a(ai)
            a._attach_to_graph(mol, n)

            charges[n] = ((hcr >> 1) & 0x0f) - 4
            radicals[n] = bool(hcr & 0x01)
            hydrogens[n] = None if (h := hcr >> 5) == 7 else h
            plane[n] = (x, y)

        bc = sum(neighbors.values()) // 2
        shift += 9 * len(neighbors) - 1
        connections = []
        for o in range(bc):
            nn = unpack_from('>I', data, shift + 3 * o)[0]
            connections.append((nn >> 12) & 0x0fff)
            connections.append(nn & 0x0fff)

        shift += 1 + 3 * bc
        orders = []
        for o in range(ceil(bc / 5)):
            bb = unpack_from('>H', data, shift + 2 * o)[0]
            orders.append(((bb >> 12) & 0x07) + 1)
            orders.append(((bb >> 9) & 0x07) + 1)
            orders.append(((bb >> 6) & 0x07) + 1)
            orders.append(((bb >> 3) & 0x07) + 1)
            orders.append((bb & 0x07) + 1)
        orders = orders[:bc]  # skip padding

        con = iter(connections)
        ords = iter(orders)
        for n, ms in neighbors.items():
            bonds[n] = cbn = {}
            for _ in range(ms):
                m = next(con)
                if m in bonds:  # bond partially exists. need back-connection.
                    cbn[m] = bonds[m][n]
                else:
                    cbn[m] = Bond(next(ords))

        shift += 2 * ceil(bc / 5)
        for o in range(acs & 0x0fff):  # cis/trans
            ct = unpack_from('>I', data, shift + 4 * o)[0]
            cis_trans_stereo[(ct >> 20, (ct >> 8) & 0x0fff)] = ct & 0x01
        return mol

    def _augmented_substructure(self, atoms: Iterable[int], deep: int):
        atoms = set(atoms)
        bonds = self._bonds
        if atoms - self._atoms.keys():
            raise ValueError('invalid atom numbers')
        nodes = [atoms]
        for _ in range(deep):
            n = {y for x in nodes[-1] for y in bonds[x]} | nodes[-1]
            if n in nodes:
                break
            nodes.append(n)
        return nodes

    @cached_property
    def _screen_fingerprint(self) -> Dict[int, Set[int]]:
        """
        Fingerprint of available linear fragments with set of mapped atoms.
        Required for isomorphism tests filtering speedup.
        Parameters can be modified globally in `MoleculeContainer._fingerprint_config`.
        """
        if self._fingerprint_config:
            return {hash(k): {x for x in v for x in x} for k, v in self._fragments(**self._fingerprint_config).items()}
        return {}

    @cached_args_method
    def _component_fingerprint(self, component):
        """
        Fingerprint of specific component.
        """
        scope = set(self.connected_components[component])
        return {k: v & scope for k, v in self._screen_fingerprint.items() if not v.isdisjoint(scope)}

    @cached_property
    def _cython_compiled_structure(self):
        # long I:
        # bond: single, double, triple, aromatic, special = 5 bit
        # atom: H-Ce: 58 bit
        # transfer bit

        # long II:
        # atom Pr-Og: 60 bit
        # hybridizations: 1-4 = 4 bit

        # long III:
        # isotope: not specified, isotope - common_isotope = -8 - +8 = 18 bit
        # is_radical: 2 bit
        # charge: -4 - +4: 9 bit
        # implicit_hydrogens: 0-4 = 5 bit
        # neighbors: 0-14 = 15 bit
        # heteroatoms: 0-14 = 15 bit

        # long IV:
        # ring_sizes: not-in-ring bit, 3-atom ring, 4-...., 65-atom ring
        from ..files._mdl.mol import common_isotopes

        charges = self._charges
        radicals = self._radicals
        hydrogens = self._hydrogens
        neighbors = self.neighbors
        heteroatoms = self.heteroatoms
        rings_sizes = self.atoms_rings_sizes
        hybridization = self.hybridization

        mapping = {}
        numbers = []
        bits1 = []
        bits2 = []
        bits3 = []
        bits4 = []
        for i, (n, a) in enumerate(self._atoms.items()):
            mapping[n] = i
            numbers.append(n)
            v2 = 1 << (hybridization(n) - 1)
            if (an := a.atomic_number) > 58:
                v1 = 1  # transfer bit
                v2 |= 1 << (122 - an)
            else:
                v1 = 1 << (59 - an)

            if a.isotope:
                v3 = 1 << (a.isotope - common_isotopes[a.atomic_symbol] + 54)
                if radicals[n]:
                    v3 |= 0x200000000000
                else:
                    v3 |= 0x100000000000
            elif radicals[n]:
                v3 = 0x8000200000000000
            else:
                v3 = 0x8000100000000000

            v3 |= 1 << (charges[n] + 39)
            v3 |= 1 << ((hydrogens[n] or 0) + 30)
            v3 |= 1 << (neighbors(n) + 15)
            v3 |= 1 << heteroatoms(n)

            if n in rings_sizes:
                v4 = 0
                for r in rings_sizes[n]:
                    if r > 65:  # big rings not supported
                        continue
                    v4 |= 1 << (65 - r)
                if not v4:  # only 65+ rings. set as rings-free.
                    v4 = 0x8000000000000000
            else:  # not in rings
                v4 = 0x8000000000000000

            bits1.append(v1)
            bits2.append(v2)
            bits3.append(v3)
            bits4.append(v4)

        o_from = [0] * len(mapping)
        o_to = [0] * len(mapping)
        indices = [0] * self.bonds_count * 2
        bonds = [0] * self.bonds_count * 2
        start = 0
        for n, ms in self._bonds.items():
            i = mapping[n]
            o_from[i] = start
            for j, (m, b) in enumerate(ms.items(), start):
                indices[j] = x = mapping[m]
                v = bits1[x]
                o = b.order
                if o == 1:
                    v |= 0x0800000000000000
                elif o == 4:
                    v |= 0x4000000000000000
                elif o == 2:
                    v |= 0x1000000000000000
                elif o == 3:
                    v |= 0x2000000000000000
                else:
                    v |= 0x8000000000000000
                bonds[j] = v
            start += len(ms)
            o_to[i] = start

        return (array('L', numbers), array('Q', bits1), array('Q', bits2), array('Q', bits3), array('Q', bits4),
                array('Q', bonds), array('I', o_from), array('I', o_to), array('I', indices))

    def _calc_implicit(self, n: int):
        atoms = self._atoms
        atom = atoms[n]
        if atom.atomic_number != 1:
            charge: int = self._charges[n]
            is_radical = self._radicals[n]
            explicit_sum = 0
            explicit_dict = defaultdict(int)
            for m, bond in self._bonds[n].items():
                order = bond.order
                if order == 4:  # aromatic rings not supported
                    self._hydrogens[n] = None
                    return
                elif order != 8:  # any bond used for complexes
                    explicit_sum += order
                    explicit_dict[(order, atoms[m].atomic_number)] += 1
            try:
                rules = atom.valence_rules(charge, is_radical, explicit_sum)
            except ValenceError:
                self._hydrogens[n] = None
                return
            for s, d, h in rules:
                if s.issubset(explicit_dict) and all(explicit_dict[k] >= c for k, c in d.items()):
                    self._hydrogens[n] = h
                    return
        self._hydrogens[n] = 0

    def __int__(self):
        """
        Total charge of molecule
        """
        return self.molecular_charge

    def __float__(self):
        return self.molecular_mass

    def __xor__(self, other):
        """
        G ^ H is CGR generation
        """
        return self.compose(other)

    def __and__(self, other: Iterable[int]):
        """
        Substructure of graph with given nodes.
        """
        return self.substructure(other)

    def __sub__(self, other: Iterable[int]):
        """
        Given nodes excluded substructure of graph.
        """
        atoms = set(other)
        if atoms - self._atoms.keys():
            raise ValueError('invalid atom numbers')
        atoms = self._atoms.keys() - atoms
        if atoms:
            return self.substructure(atoms)
        raise ValueError('full substitution not allowed')

    def __enter__(self):
        """
        Transaction of changes. Keep current state for restoring on errors.
        """
        atoms = {}
        for n, atom in self._atoms.items():
            atom = atom.copy()
            atoms[n] = atom
            atom._attach_to_graph(self, n)

        bonds = {}
        for n, m_bond in self._bonds.items():
            bonds[n] = cbn = {}
            for m, bond in m_bond.items():
                if m in bonds:  # bond partially exists. need back-connection.
                    cbn[m] = bonds[m][n]
                else:
                    cbn[m] = bond.copy()
        self._backup = {'atoms': atoms, 'bonds': bonds, 'parsed_mapping': self._parsed_mapping.copy(),
                        'plane': self._plane.copy(), 'charges': self._charges.copy(), 'radicals': self._radicals.copy(),
                        'hydrogens': self._hydrogens.copy(), 'conformers': [x.copy() for x in self._conformers],
                        'atoms_stereo': self._atoms_stereo.copy(), 'allenes_stereo': self._allenes_stereo.copy(),
                        'cis_trans_stereo': self._cis_trans_stereo.copy()}
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:  # restore state
            backup = self._backup
            self._atoms = backup['atoms']
            self._bonds = backup['bonds']
            self._parsed_mapping = backup['parsed_mapping']
            self._plane = backup['plane']
            self._charges = backup['charges']
            self._radicals = backup['radicals']
            self._hydrogens = backup['hydrogens']
            self._conformers = backup['conformers']
            self._atoms_stereo = backup['atoms_stereo']
            self._allenes_stereo = backup['allenes_stereo']
            self._cis_trans_stereo = backup['cis_trans_stereo']
            self.flush_cache()
        del self._backup

    def __getstate__(self):
        return {'conformers': self._conformers, 'hydrogens': self._hydrogens, 'atoms_stereo': self._atoms_stereo,
                'allenes_stereo': self._allenes_stereo, 'cis_trans_stereo': self._cis_trans_stereo,
                'parsed_mapping': self._parsed_mapping, 'meta': self.__meta, 'name': self.__name,
                **super().__getstate__()}

    def __setstate__(self, state):
        super().__setstate__(state)
        self._conformers = state['conformers']
        self._atoms_stereo = state['atoms_stereo']
        self._allenes_stereo = state['allenes_stereo']
        self._cis_trans_stereo = state['cis_trans_stereo']
        self._hydrogens = state['hydrogens']
        self._parsed_mapping = state['parsed_mapping']
        self.__meta = state['meta']
        self.__name = state['name']

    _fingerprint_config = {'min_radius': 2, 'max_radius': 4}  # set empty for disable screening


__all__ = ['MoleculeContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict, deque
from functools import cached_property
from itertools import combinations
from operator import itemgetter
from typing import Any, Dict, List, Optional, Set, Tuple, TYPE_CHECKING, Union
from ..exceptions import ImplementationError


if TYPE_CHECKING:
    from chython.containers.graph import Graph


class Rings:
    __slots__ = ()

    @cached_property
    def sssr(self) -> Tuple[Tuple[int, ...], ...]:
        """
        Smallest Set of Smallest Rings. Special bonds ignored.

        Based on idea of PID matrices from:
        Lee, C. J., Kang, Y.-M., Cho, K.-H., & No, K. T. (2009).
        A robust method for searching the smallest set of smallest rings with a path-included distance matrix.
        Proceedings of the National Academy of Sciences of the United States of America, 106(41), 17355–17358.
        https://doi.org/10.1073/pnas.0813040106

        :return rings atoms numbers
        """
        if self.rings_count:
            return _sssr(self.not_special_connectivity, self.rings_count)
        return ()

    @cached_property
    def aromatic_rings(self: 'Graph') -> Tuple[Tuple[int, ...], ...]:
        """
        Aromatic rings atoms numbers
        """
        bonds = self._bonds
        return tuple(ring for ring in self.sssr if bonds[ring[0]][ring[-1]] == 4
                     and all(bonds[n][m] == 4 for n, m in zip(ring, ring[1:])))

    @cached_property
    def atoms_rings(self) -> Dict[int, Tuple[Tuple[int, ...]]]:
        """
        Dict of atoms rings which contains it.
        """
        rings = defaultdict(list)
        for r in self.sssr:
            for n in r:
                rings[n].append(r)
        return {n: tuple(rs) for n, rs in rings.items()}

    @cached_property
    def atoms_rings_sizes(self) -> Dict[int, Tuple[int, ...]]:
        """
        Sizes of rings containing atom.
        """
        return {n: tuple(len(r) for r in rs) for n, rs in self.atoms_rings.items()}

    @cached_property
    def ring_atoms(self):
        """
        Atoms in rings. Not SSSR based fast algorithm.
        """
        bonds = _skin_graph(self.not_special_connectivity)
        if not bonds:
            return set()

        in_rings = set()
        atoms = set(bonds)
        while atoms:
            stack = deque([(atoms.pop(), 0, 0)])
            path = []
            seen = set()
            while stack:
                c, p, d = stack.pop()
                if len(path) > d:
                    path = path[:d]
                if c in in_rings:
                    continue
                path.append(c)
                seen.add(c)

                d += 1
                for n in bonds[c]:
                    if n == p:
                        continue
                    elif n in seen:
                        in_rings.update(path[path.index(n):])
                    else:
                        stack.append((n, c, d))

            atoms.difference_update(seen)
        return in_rings

    @cached_property
    def rings_count(self) -> int:
        """
        SSSR rings count. Ignored rings with special bonds.
        """
        bonds = self.not_special_connectivity
        return sum(len(x) for x in bonds.values()) // 2 - len(bonds) + len(_connected_components(bonds))

    @cached_property
    def not_special_connectivity(self: 'Graph') -> Dict[int, Set[int]]:
        """
        Graph connectivity without special bonds.
        """
        bonds = {}
        for n, ms in self._bonds.items():
            bonds[n] = ngb = set()
            for m, b in ms.items():
                if b != 8:
                    ngb.add(m)
        return bonds

    @cached_property
    def connected_components(self: 'Graph') -> Tuple[Tuple[int, ...], ...]:
        """
        Isolated components of single graph. E.g. salts as ion pair.
        """
        if not self._atoms:
            return ()
        return tuple(tuple(x) for x in _connected_components(self._bonds))

    @property
    def connected_components_count(self) -> int:
        """
        Number of components in graph
        """
        return len(self.connected_components)

    @cached_property
    def skin_graph(self: 'Graph') -> Dict[int, Set[int]]:
        """
        Graph without terminal atoms. Only rings and linkers
        """
        return _skin_graph(self._bonds)


def _sssr(bonds: Dict[int, Union[Set[int], Dict[int, Any]]], n_sssr: int) -> Tuple[Tuple[int, ...], ...]:
    """
    Smallest Set of Smallest Rings of any adjacency matrix.
    Number of rings required.
    """
    bonds = _skin_graph(bonds)
    paths = _bfs(bonds)
    pid1, pid2, dist = _make_pid(paths)
    return _rings_filter(_c_set(pid1, pid2, dist), n_sssr)


def _connected_components(bonds: Dict[int, Union[Set[int], Dict[int, Any]]]) -> List[Set[int]]:
    atoms = set(bonds)
    components = []
    while atoms:
        start = atoms.pop()
        seen = {start}
        queue = deque([start])
        while queue:
            current = queue.popleft()
            for i in bonds[current]:
                if i not in seen:
                    queue.append(i)
                    seen.add(i)
        components.append(seen)
        atoms.difference_update(seen)
    return components


def _skin_graph(bonds: Dict[int, Union[Set[int], Dict[int, Any]]]) -> Dict[int, Set[int]]:
    """
    Graph without terminal nodes. Only rings and linkers
    """
    bonds = {n: set(ms) for n, ms in bonds.items() if ms}
    while True:  # skip not-cycle chains
        try:
            n = next(n for n, ms in bonds.items() if len(ms) <= 1)
        except StopIteration:
            break
        for m in bonds.pop(n):
            bonds[m].discard(n)
    return bonds


def _bfs(bonds):
    atoms = set(bonds)
    terminated = []
    tail = atoms.pop()
    next_stack = {x: [tail, x] for x in bonds[tail]}

    while True:
        next_front = set()
        found_odd = set()
        stack, next_stack = next_stack, {}
        for tail, path in stack.items():
            neighbors = bonds[tail] & atoms
            next_front.add(tail)

            if len(neighbors) == 1:
                n = neighbors.pop()
                if n in found_odd:
                    if len(path) != 1:
                        terminated.append(tuple(path))  # save second ring closure
                    next_stack[n] = [n]  # maybe we have another path?
                else:
                    path.append(n)
                    if n in stack:  # odd rings
                        found_odd.add(tail)
                        terminated.append(tuple(path))  # found ring closure. save path.
                    elif n in next_stack:  # even rings
                        terminated.append(tuple(path))
                        if len(next_stack[n]) != 1:  # prevent bicycle case
                            terminated.append(tuple(next_stack[n]))
                            next_stack[n] = [n]
                    else:
                        next_stack[n] = path  # grow must go on
            elif neighbors:
                if len(path) != 1:
                    terminated.append(tuple(path))  # save path.
                for n in neighbors:
                    if n in found_odd:
                        if n in stack:
                            if n in next_stack:
                                del next_stack[n]
                        else:
                            next_stack[n] = [n]
                    else:
                        path = [tail, n]
                        if n in stack:  # odd rings
                            found_odd.add(tail)
                            terminated.append(tuple(path))
                        elif n in next_stack:  # even rings
                            terminated.append(tuple(path))
                            if len(next_stack[n]) != 1:  # prevent bicycle case
                                terminated.append(tuple(next_stack[n]))
                                next_stack[n] = [n]
                        else:
                            next_stack[n] = path

        atoms.difference_update(next_front)
        if not atoms:
            break
        elif not next_stack:
            tail = atoms.pop()
            next_stack = {x: [tail, x] for x in bonds[tail] & atoms}
    return terminated


def _make_pid(paths: List[List[int]]):
    pid1 = defaultdict(lambda: defaultdict(dict))
    pid2 = defaultdict(lambda: defaultdict(dict))
    distances = defaultdict(lambda: defaultdict(lambda: 1e9))
    chains = sorted(paths, key=len)
    for c in chains:
        di = len(c) - 1
        n, m = c[0], c[-1]
        nn, mm = c[1], c[-2]
        if n in distances and m in distances[n] and distances[n][m] != di:
            pid2[n][m][(nn, mm)] = c
            pid2[m][n][(mm, nn)] = c[::-1]
        else:
            pid1[n][m][(nn, mm)] = c
            pid1[m][n][(mm, nn)] = c[::-1]
            distances[n][m] = distances[m][n] = di

    for k in pid1:
        new_distances = defaultdict(dict)
        dk = distances[k]
        ndk = new_distances[k]
        for i in pid1:
            if i == k:
                continue
            di = distances[i]
            ndi = new_distances[i]
            ndk[i] = ndi[k] = di[k]
            for j in pid1:
                if j == k or j == i:
                    continue
                ij = di[j]
                ikj = di[k] + dk[j]
                if ij - ikj == 1:  # A new shortest path == previous shortest path - 1
                    pid2[i][j] = pid1[i][j]
                    pid1[i][j] = {(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                  zip(pid1[i][k].items(), pid1[k][j].items())}
                    ndi[j] = ikj
                elif ij > ikj:  # A new shortest path
                    pid2[i][j] = {}
                    pid1[i][j] = {(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                  zip(pid1[i][k].items(), pid1[k][j].items())}
                    ndi[j] = ikj
                elif ij == ikj:  # Another shortest path
                    pid1[i][j].update({(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                       zip(pid1[i][k].items(), pid1[k][j].items())})
                    ndi[j] = ij
                elif ikj - ij == 1:  # Shortest+1 path
                    pid2[i][j].update({(ni, mj): ip[:-1] + jp for ((ni, _), ip), ((_, mj), jp) in
                                       zip(pid1[i][k].items(), pid1[k][j].items())})
                    ndi[j] = ij
                else:
                    ndi[j] = ij
        distances = new_distances
    return pid1, pid2, distances


def _c_set(pid1, pid2, pid1l):
    c_set = []
    seen = set()
    for i, p1i in pid1.items():
        seen.add(i)
        di = pid1l[i]
        p2i = pid2[i]

        for j, p1ij in p1i.items():
            if j in seen:
                continue
            p1ij = list(p1ij.values())
            p2ij = list(p2i[j].values())
            dij = di[j] * 2

            if len(p1ij) == 1:  # one shortest
                if not p2ij:  # need shortest + 1 path
                    continue
                c_set.append((dij + 1, p1ij, p2ij))
            elif not p2ij:  # one or more odd rings
                c_set.append((dij, p1ij, None))
            else:  # odd and even rings found (e.g. bicycle)
                c_set.append((dij, p1ij, None))
                c_set.append((dij + 1, p1ij, p2ij))

    for c_num, p1ij, p2ij in sorted(c_set, key=itemgetter(0)):
        if c_num % 2:  # odd rings
            for c1 in p1ij:
                for c2 in p2ij:
                    c = c1 + c2[-2:0:-1]
                    if len(c) == len(set(c)):
                        yield _canonic_ring(c)
        else:
            for c1, c2 in zip(p1ij, p1ij[1:]):
                c = c1 + c2[-2:0:-1]
                if len(c) == len(set(c)):
                    yield _canonic_ring(c)


def _canonic_ring(ring: Tuple[int, ...]) -> Tuple[int, ...]:
    n = min(ring)
    ndx = ring.index(n)
    if ndx == 0:
        if ring[-1] < ring[1]:
            return n, *ring[:0:-1]
        return ring
    elif ndx == len(ring) - 1:
        if ring[0] > ring[-2]:
            return ring[::-1]
        return n, *ring[:-1]
    if ring[ndx + 1] > ring[ndx - 1]:
        return *ring[ndx::-1], *ring[:ndx:-1]
    return *ring[ndx:], *ring[:ndx]


def _ring_scissors(ring: Tuple[int, ...], n: int, m: int) -> Tuple[int, ...]:
    ndx = ring.index(n)
    mdx = ring.index(m)
    if ndx == 0:
        if mdx == 1:
            return n, *ring[:0:-1]
        return ring
    elif ndx == len(ring) - 1:
        if mdx == 0:
            return ring[::-1]
        return n, *ring[:-1]
    if ndx < mdx:
        return *ring[ndx::-1], *ring[:ndx:-1]
    return *ring[ndx:], *ring[:ndx]


def _ring_adjacency(ring: Tuple[int, ...]) -> Dict[int, List[int]]:
    adj = {ring[0]: [ring[-1]]}  # ring adjacency matrix
    for n, m in zip(ring, ring[1:]):
        adj[n].append(m)
        adj[m] = [n]
    adj[m].append(ring[0])
    return adj


def _is_condensed_ring(c, sssr, seen_rings):
    # create graph of connected neighbour rings
    ck = seen_rings[c]
    neighbors = {x: set() for x in sssr if len(seen_rings[x].keys() & ck.keys()) > 1}
    if len(neighbors) > 1:
        for (i, iv), (j, jv) in combinations(neighbors.items(), 2):
            if len(seen_rings[i].keys() & seen_rings[j].keys()) > 1:
                iv.add(j)
                jv.add(i)
        # check if hold rings is combination of existing. (123654) is combo of (1254) and (2365)
        #
        # 1--2--3
        # |  |  |
        # 4--5--6
        #
        # modified NX.dfs_labeled_edges
        # https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.\
        # traversal.depth_first_search.dfs_labeled_edges.html
        depth_limit = len(neighbors) - 1
        for start, nbrs in neighbors.items():
            if not nbrs:
                continue
            stack = [(start, seen_rings[start], depth_limit, iter(nbrs), {start})]
            while stack:
                parent, p_adj, depth_now, children, seen = stack[-1]
                try:
                    child = next(children)
                except StopIteration:
                    stack.pop()
                else:
                    if child not in seen:
                        common = p_adj.keys() & seen_rings[child].keys()
                        if len(common) > 2:  # only terminal common atoms required
                            term = {n for n in common if len(common.intersection(p_adj[n])) == 1}
                            if len(term) != 2:  # skip multiple contacts
                                continue
                            common.difference_update(term)
                            n, m = term
                            mc = _canonic_ring(
                                    (*_ring_scissors(tuple(x for x in parent if x not in common), n, m),
                                     *_ring_scissors(tuple(x for x in child if x not in common), m, n)[1:-1]))
                        elif len(common) == 2:
                            n, m = common
                            mc = _canonic_ring((*_ring_scissors(parent, n, m), *_ring_scissors(child, m, n)[1:-1]))
                        else:  # point connections
                            continue
                        if c == mc:  # macrocycle found
                            return True
                        elif depth_now and 2 < len(mc) <= len(c) + 1:
                            stack.append((mc, _ring_adjacency(mc), depth_now - 1, iter(neighbors[child]),
                                          {child} | seen))
    return False


def _get_unique_chord(ring: Tuple[int, ...], common: Set[int]) -> Optional[Tuple[int, ...]]:
    lc = len(common)
    if len(ring) == lc:
        if common == set(ring):
            return ()
    else:
        if common == set(ring[:lc]):
            return *ring[lc - 1:], ring[0]
        for _ in range(len(ring) - 1):
            ring = (*ring[1:], ring[0])
            if common == set(ring[:lc]):
                return *ring[lc - 1:], ring[0]


def _connected_rings(rings, seen_rings):
    rings = rings.copy()
    out = []
    for i in range(len(rings)):
        c = rings[i]
        ck = seen_rings[c]
        for j in range(i + 1, len(rings)):
            r = rings[j]
            rk = seen_rings[r]
            common = rk.keys() & ck.keys()
            if len(common) == 2:  # one common bond
                n, m = common
                if m in ck[n] and m in rk[n]:  # only common bond!
                    c = _canonic_ring((*_ring_scissors(c, n, m), *_ring_scissors(r, m, n)[1:-1]))
                    ck = _ring_adjacency(c)
                    rings[j] = c
                    seen_rings[c] = ck
                    break
            elif len(common) > 2:
                cc = _get_unique_chord(c, common)
                if cc is None:  # skip multitouched rings
                    continue
                r = _get_unique_chord(r, common)
                if r is None:
                    continue
                if cc:
                    if r:
                        if r[0] == cc[0]:
                            r = r[::-1]
                        c = _canonic_ring((*cc, *r[1:-1]))
                        ck = _ring_adjacency(c)
                        rings[j] = c
                        seen_rings[c] = ck
                        break
                    else:
                        c = _canonic_ring(cc)
                        ck = _ring_adjacency(c)
                        rings[j] = c
                        seen_rings[c] = ck
                        break
                elif r:
                    c = _canonic_ring(r)
                    ck = _ring_adjacency(c)
                    rings[j] = c
                    seen_rings[c] = ck
                    break
        else:  # isolated ring[s] found
            out.append(c)
    return out


def _rings_filter(rings, n_sssr):
    c = next(rings)
    if n_sssr == 1:
        return c,

    seen_rings = {c}
    sssr_atoms = set(c)
    sssr = [c]
    hold = []
    for c in rings:
        if c in seen_rings:
            continue
        seen_rings.add(c)
        if sssr_atoms.issuperset(c):  # potentially condensed ring
            hold.append(c)
            continue
        sssr_atoms.update(c)
        sssr.append(c)
        if len(sssr) == n_sssr:
            return tuple(sssr)

    # now we have set of plug rings (cuban fullerene), besiege rings and condensed trash
    seen_rings = {c: _ring_adjacency(c) for c in seen_rings}  # prepare adjacency
    condensed_rings = _connected_rings(sssr, seen_rings)  # collection of contours of condensed rings

    for c in hold:
        if c in condensed_rings or _is_condensed_ring(c, sssr, seen_rings):
            continue
        condensed_rings.insert(0, c)
        condensed_rings = _connected_rings(condensed_rings, seen_rings)
        sssr.append(c)
        if len(sssr) == n_sssr:
            return tuple(sorted(sssr, key=len))

    raise ImplementationError('SSSR count not reached')


__all__ = ['Rings']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Dmitrij Zanadvornykh <zandmitrij@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC
from typing import Tuple, Type, List, Union
from .core import Core
from .element import Element
from ...exceptions import IsNotConnectedAtom


_inorganic = {'He', 'Ne', 'Ar', 'Kr', 'Xe', 'F', 'Cl', 'Br', 'I', 'B', 'C', 'N', 'O',
              'H', 'Si', 'P', 'S', 'Se', 'Ge', 'As', 'Sb', 'Te'}


class Query(Core, ABC):
    __slots__ = ()

    @Core.charge.setter
    def charge(self, charge):
        if not isinstance(charge, int):
            raise TypeError('formal charge should be int in range [-4, 4]')
        elif charge > 4 or charge < -4:
            raise ValueError('formal charge should be in range [-4, 4]')
        try:
            g = self._graph()
            g._charges[self._map] = charge
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()

    @Core.is_radical.setter
    def is_radical(self, is_radical):
        if not isinstance(is_radical, bool):
            raise TypeError('bool expected')
        try:
            g = self._graph()
            g._radicals[self._map] = is_radical
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()

    @property
    def neighbors(self) -> Tuple[int, ...]:
        try:
            return self._graph()._neighbors[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def hybridization(self):
        try:
            return self._graph()._hybridizations[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @neighbors.setter
    def neighbors(self, neighbors):
        try:
            g = self._graph()
            g._neighbors[self._map] = g._validate_neighbors(neighbors)
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()

    @hybridization.setter
    def hybridization(self, hybridization):
        try:
            g = self._graph()
            g._hybridizations[self._map] = g._validate_hybridization(hybridization)
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()

    @property
    def heteroatoms(self) -> Tuple[int, ...]:
        try:
            return self._graph()._heteroatoms[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @heteroatoms.setter
    def heteroatoms(self, heteroatoms):
        try:
            g = self._graph()
            g._heteroatoms[self._map] = g._validate_neighbors(heteroatoms)
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()

    @property
    def ring_sizes(self) -> Tuple[int, ...]:
        """
        Atom rings sizes.
        """
        try:
            return self._graph()._rings_sizes[self._map]
        except AttributeError:
            raise IsNotConnectedAtom
        except KeyError:
            return ()

    @ring_sizes.setter
    def ring_sizes(self, ring_sizes):
        try:
            g = self._graph()
            g._rings_sizes[self._map] = g._validate_rings(ring_sizes)
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()

    @property
    def implicit_hydrogens(self) -> Tuple[int, ...]:
        try:
            return self._graph()._hydrogens[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @implicit_hydrogens.setter
    def implicit_hydrogens(self, implicit_hydrogens):
        try:
            g = self._graph()
            g._hydrogens[self._map] = g._validate_neighbors(implicit_hydrogens)
        except AttributeError:
            raise IsNotConnectedAtom
        else:
            g.flush_cache()


class QueryElement(Query, ABC):
    __slots__ = ()

    @property
    def atomic_symbol(self) -> str:
        return self.__class__.__name__[5:]

    @classmethod
    def from_symbol(cls, symbol: str) -> Type[Union['QueryElement', 'AnyElement', 'AnyMetal']]:
        """
        get Element class by its symbol
        """
        if symbol == 'A':
            return AnyElement
        elif symbol == 'M':
            return AnyMetal
        try:
            element = next(x for x in QueryElement.__subclasses__() if x.__name__ == f'Query{symbol}')
        except StopIteration:
            raise ValueError(f'QueryElement with symbol "{symbol}" not found')
        return element

    @classmethod
    def from_atomic_number(cls, number: int) -> Type['QueryElement']:
        """
        get Element class by its number
        """
        try:
            element = next(x for x in QueryElement.__subclasses__() if x.atomic_number.fget(None) == number)
        except StopIteration:
            raise ValueError(f'QueryElement with number "{number}" not found')
        return element

    @classmethod
    def from_atom(cls, atom: Union['Element', 'Query']) -> 'Query':
        """
        get QueryElement or AnyElement object from Element object or copy of QueryElement or AnyElement
        """
        if isinstance(atom, Element):
            return cls.from_atomic_number(atom.atomic_number)(atom.isotope)
        elif not isinstance(atom, Query):
            raise TypeError('Element or Query expected')
        return atom.copy()

    def __eq__(self, other):
        """
        compare attached to molecules elements and query elements
        """
        if isinstance(other, Element):
            if self.atomic_number == other.atomic_number and self.charge == other.charge and \
                    self.is_radical == other.is_radical:
                if self.isotope and self.isotope != other.isotope:
                    return False
                if self.neighbors and other.neighbors not in self.neighbors:
                    return False
                if self.hybridization and other.hybridization not in self.hybridization:
                    return False
                if self.ring_sizes:
                    if self.ring_sizes[0]:
                        if set(self.ring_sizes).isdisjoint(other.ring_sizes):
                            return False
                    elif other.ring_sizes:  # not in ring expected
                        return False
                if self.implicit_hydrogens and other.implicit_hydrogens not in self.implicit_hydrogens:
                    return False
                if self.heteroatoms and other.heteroatoms not in self.heteroatoms:
                    return False
                return True
        elif isinstance(other, QueryElement) and self.atomic_number == other.atomic_number and \
                self.isotope == other.isotope and self.charge == other.charge and self.is_radical == other.is_radical \
                and self.neighbors == other.neighbors and self.hybridization == other.hybridization \
                and self.ring_sizes == other.ring_sizes and self.implicit_hydrogens == other.implicit_hydrogens \
                and self.heteroatoms == other.heteroatoms:
            # equal query element has equal query marks
            return True
        return False

    def __hash__(self):
        return hash((self.isotope or 0, self.atomic_number, self.charge, self.is_radical, self.neighbors,
                     self.hybridization, self.ring_sizes, self.implicit_hydrogens, self.heteroatoms))


class AnyElement(Query):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__()

    @property
    def atomic_symbol(self) -> str:
        return 'A'

    @property
    def atomic_number(self) -> int:
        return 0

    def __eq__(self, other):
        """
        Compare attached to molecules elements and query elements
        """
        if isinstance(other, Element):
            if self.charge == other.charge and self.is_radical == other.is_radical:
                if self.neighbors and other.neighbors not in self.neighbors:
                    return False
                if self.hybridization and other.hybridization not in self.hybridization:
                    return False
                if self.ring_sizes:
                    if self.ring_sizes[0]:
                        if set(self.ring_sizes).isdisjoint(other.ring_sizes):
                            return False
                    elif other.ring_sizes:  # not in ring expected
                        return False
                if self.implicit_hydrogens and other.implicit_hydrogens not in self.implicit_hydrogens:
                    return False
                if self.heteroatoms and other.heteroatoms not in self.heteroatoms:
                    return False
                return True
        elif isinstance(other, AnyMetal):
            return False
        elif isinstance(other, Query) and self.charge == other.charge and self.is_radical == other.is_radical \
                and self.neighbors == other.neighbors and self.hybridization == other.hybridization \
                and self.ring_sizes == other.ring_sizes and self.implicit_hydrogens == other.implicit_hydrogens \
                and self.heteroatoms == other.heteroatoms:
            return True
        return False

    def __hash__(self):
        return hash((self.charge, self.is_radical, self.neighbors, self.hybridization, self.ring_sizes,
                     self.implicit_hydrogens, self.heteroatoms))


class AnyMetal(Query):
    """
    Charge and radical ignored any metal. Rings, hydrogens and heteroatoms count also ignored.

    Class designed for d-elements matching in standardization.
    """
    def __init__(self, *args, **kwargs):
        super().__init__()

    @property
    def atomic_symbol(self) -> str:
        return 'M'

    @property
    def atomic_number(self) -> int:
        return 0

    def __eq__(self, other):
        if isinstance(other, Element):
            if other.atomic_symbol not in _inorganic:
                if self.neighbors and other.neighbors not in self.neighbors:
                    return False
                if self.hybridization and other.hybridization not in self.hybridization:
                    return False
                return True
        elif isinstance(other, AnyMetal) and self.neighbors == other.neighbors \
                and self.hybridization == other.hybridization:
            return True
        return False

    def __hash__(self):
        return hash((self.neighbors, self.hybridization))


class ListElement(Query):
    __slots__ = ('_elements', '_numbers')

    def __init__(self, elements: List[str], *args, **kwargs):
        """
        Elements list
        """
        super().__init__()
        self._elements = tuple(elements)
        self._numbers = tuple(x.atomic_number.fget(None) for x in Element.__subclasses__() if x.__name__ in elements)

    @property
    def atomic_symbol(self) -> str:
        return ','.join(self._elements)

    @property
    def atomic_number(self) -> int:
        return 0

    def copy(self):
        copy = super().copy()
        copy._elements = self._elements
        copy._numbers = self._numbers
        return copy

    def __eq__(self, other):
        """
        Compare attached to molecules elements and query elements
        """
        if isinstance(other, Element):
            if other.atomic_number in self._numbers:
                if self.charge != other.charge or self.is_radical != other.is_radical:
                    return False
                if self.neighbors and other.neighbors not in self.neighbors:
                    return False
                if self.hybridization and other.hybridization not in self.hybridization:
                    return False
                if self.ring_sizes:
                    if self.ring_sizes[0]:
                        if set(self.ring_sizes).isdisjoint(other.ring_sizes):
                            return False
                    elif other.ring_sizes:  # not in ring expected
                        return False
                if self.implicit_hydrogens and other.implicit_hydrogens not in self.implicit_hydrogens:
                    return False
                if self.heteroatoms and other.heteroatoms not in self.heteroatoms:
                    return False
                return True
        elif isinstance(other, (AnyElement, AnyMetal)):
            return False
        elif isinstance(other, Query) and self.charge == other.charge and self.is_radical == other.is_radical \
                and self.neighbors == other.neighbors and self.hybridization == other.hybridization \
                and self.ring_sizes == other.ring_sizes and self.implicit_hydrogens == other.implicit_hydrogens \
                and self.heteroatoms == other.heteroatoms:
            if isinstance(other, ListElement):
                return self._numbers == other._numbers
            return other.atomic_number in self._numbers
        return False

    def __hash__(self):
        return hash((self._numbers, self.charge, self.is_radical, self.neighbors, self.hybridization,
                     self.ring_sizes, self.implicit_hydrogens, self.heteroatoms))

    def __getstate__(self):
        state = super().__getstate__()
        state['elements'] = self._elements
        return state

    def __setstate__(self, state):
        self._elements = state['elements']
        self._numbers = tuple(x.atomic_number.fget(None) for x in Element.__subclasses__()
                              if x.__name__ in state['elements'])
        super().__setstate__(state)

    def __repr__(self):
        return f'{self.__class__.__name__}([{",".join(self._elements)}])'


__all__ = ['Query', 'QueryElement', 'AnyElement', 'AnyMetal', 'ListElement']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Adelia Fatykhova <adelik21979@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import Union
from .base import BaseReactor
from ..containers import QueryContainer, MoleculeContainer


class Transformer(BaseReactor):
    """
    Editor for molecules.
    generates modified molecules from input molecule using template.
    Transformer calling returns generator of all possible replacements.
    """
    def __init__(self, pattern: QueryContainer, replacement: Union[MoleculeContainer, QueryContainer],
                 delete_atoms: bool = True):
        """
        :param pattern: search pattern
        :param replacement: atoms and bonds replacement
        :param delete_atoms: if True atoms exists in pattern but not exists in replacement will be removed
        """
        if not isinstance(pattern, QueryContainer) or not isinstance(replacement, (MoleculeContainer, QueryContainer)):
            raise TypeError('invalid params')

        self.__pattern = pattern
        super().__init__(pattern, replacement, delete_atoms)

    def __call__(self, structure: MoleculeContainer, automorphism_filter: bool = True):
        if not isinstance(structure, MoleculeContainer):
            raise TypeError('only Molecules possible')

        for mapping in self.__pattern.get_mapping(structure, automorphism_filter=automorphism_filter):
            yield self._patcher(structure, mapping)

    def __getstate__(self):
        return {'pattern': self.__pattern, **super().__getstate__()}

    def __setstate__(self, state):
        self.__pattern = state['pattern']
        super().__setstate__(state)


__all__ = ['Transformer']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from ._write import _MDLWrite
from ...containers import MoleculeContainer


charge_map = {-4: '  0', -3: '  7', -2: '  6', -1: '  5', 0: '  0', 1: '  3', 2: '  2', 3: '  1', 4: '  0'}


class MDLWrite(_MDLWrite):
    def _convert_molecule(self, g, write3d=None):
        if not isinstance(g, MoleculeContainer):
            raise TypeError('MoleculeContainer expected')
        elif max(g) > 999:
            raise ValueError('MOL file support only small molecules')

        out = [f'{g.name}\n\n\n{g.atoms_count:3d}{g.bonds_count:3d}  0  0  0  0            999 V2000\n']
        if write3d is not None:
            out.append(self.__convert_atoms3d(g, g._conformers[write3d]))
        else:
            out.append(self.__convert_atoms2d(g))
        out.append(self.__convert_bonds(g))
        gc = g._charges
        gr = g._radicals
        for n, (m, a) in enumerate(g._atoms.items(), start=1):
            if a.isotope:
                out.append(f'M  ISO  1 {n:3d} {a.isotope:3d}\n')
            if gr[m]:
                out.append(f'M  RAD  1 {n:3d}   2\n')  # invalid for carbenes
            c = gc[m]
            if c in (-4, 4):
                out.append(f'M  CHG  1 {n:3d} {c:3d}\n')
        out.append('M  END\n')
        return ''.join(out)

    def __convert_atoms2d(self, g):
        gc = g._charges
        gp = g._plane
        out = []
        for n, (m, a) in enumerate(g._atoms.items(), start=1):
            x, y = gp[m]
            c = gc[m]
            if c in (-4, 4):
                if self._mapping:
                    out.append(f'{x:10.4f}{y:10.4f}    0.0000 {a.atomic_symbol:3s} '
                               f'0  0  0  0  0  0  0  0  0{m:3d}  0  0\n')
                else:
                    out.append(f'{x:10.4f}{y:10.4f}    0.0000 {a.atomic_symbol:3s} '
                               f'0  0  0  0  0  0  0  0  0  0  0  0\n')
            else:
                if self._mapping:
                    out.append(f'{x:10.4f}{y:10.4f}    0.0000 {a.atomic_symbol:3s} 0{charge_map[c]}  0  0  0  0'
                               f'  0  0  0{m:3d}  0  0\n')
                else:
                    out.append(f'{x:10.4f}{y:10.4f}    0.0000 {a.atomic_symbol:3s} 0{charge_map[c]}  0  0  0  0'
                               f'  0  0  0  0  0  0\n')
        return ''.join(out)

    def __convert_atoms3d(self, g, xyz):
        gc = g._charges
        out = []
        for n, (m, a) in enumerate(g._atoms.items(), start=1):
            x, y, z = xyz[m]
            c = gc[m]
            if c in (-4, 4):
                if self._mapping:
                    out.append(f'{x:10.4f}{y:10.4f}{z:10.4f} {a.atomic_symbol:3s} '
                               f'0  0  0  0  0  0  0  0  0{m:3d}  0  0\n')
                else:
                    out.append(f'{x:10.4f}{y:10.4f}{z:10.4f} {a.atomic_symbol:3s} 0  0  0  0  0  0  0  0  0  0  0  0\n')
            else:
                if self._mapping:
                    out.append(f'{x:10.4f}{y:10.4f}{z:10.4f} {a.atomic_symbol:3s} 0{charge_map[c]}  0  0  0  0'
                               f'  0  0  0{m:3d}  0  0\n')
                else:
                    out.append(f'{x:10.4f}{y:10.4f}{z:10.4f} {a.atomic_symbol:3s} 0{charge_map[c]}  0  0  0  0'
                               f'  0  0  0  0  0  0\n')
        return ''.join(out)

    @staticmethod
    def __convert_bonds(g):
        bonds = g._bonds
        atoms = {m: n for n, m in enumerate(g._atoms, start=1)}
        wedge = defaultdict(set)
        out = []
        for n, m, s in g._wedge_map:
            out.append(f'{atoms[n]:3d}{atoms[m]:3d}  {bonds[n][m].order}  {s == 1 and "1" or "6"}  0  0  0\n')
            wedge[n].add(m)
            wedge[m].add(n)
        for n, m, b in g.bonds():
            if m not in wedge[n]:
                out.append(f'{atoms[n]:3d}{atoms[m]:3d}  {b.order}  0  0  0  0\n')
        return ''.join(out)


__all__ = ['MDLWrite']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC
from typing import Type, Union
from .core import Core
from .element import Element
from ...exceptions import IsNotConnectedAtom


class DynamicElement(Core, ABC):
    __slots__ = ('__p_charge', '__p_is_radical')

    @property
    def atomic_symbol(self) -> str:
        return self.__class__.__name__[7:]

    @classmethod
    def from_symbol(cls, symbol: str) -> Type['DynamicElement']:
        """
        get DynamicElement class by its symbol
        """
        try:
            element = next(x for x in DynamicElement.__subclasses__() if x.__name__ == f'Dynamic{symbol}')
        except StopIteration:
            raise ValueError(f'DynamicElement with symbol "{symbol}" not found')
        return element

    @classmethod
    def from_atomic_number(cls, number: int) -> Type['DynamicElement']:
        """
        get DynamicElement class by its number
        """
        try:
            element = next(x for x in DynamicElement.__subclasses__() if x.atomic_number.fget(None) == number)
        except StopIteration:
            raise ValueError(f'DynamicElement with number "{number}" not found')
        return element

    @classmethod
    def from_atom(cls, atom: Union['Element', 'DynamicElement']) -> 'DynamicElement':
        """
        get DynamicElement object from Element object or copy of DynamicElement object
        """
        if isinstance(atom, Element):
            return cls.from_atomic_number(atom.atomic_number)(atom.isotope)
        elif not isinstance(atom, DynamicElement):
            raise TypeError('Element or DynamicElement expected')
        return atom.copy()

    @property
    def p_charge(self) -> int:
        try:
            return self._graph()._p_charges[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    @property
    def p_is_radical(self) -> bool:
        try:
            return self._graph()._p_radicals[self._map]
        except AttributeError:
            raise IsNotConnectedAtom

    def __eq__(self, other):
        """
        compare attached to molecules dynamic elements
        """
        return isinstance(other, DynamicElement) and self.atomic_number == other.atomic_number and \
            self.isotope == other.isotope and self.charge == other.charge and self.is_radical == other.is_radical and \
            self.p_charge == other.p_charge and self.p_is_radical == other.p_is_radical

    def __hash__(self):
        return hash((self.isotope or 0, self.atomic_number, self.charge, self.p_charge,
                     self.is_radical, self.p_is_radical))

    @property
    def is_dynamic(self) -> bool:
        """
        Atom has dynamic features
        """
        return self.charge != self.p_charge or self.is_radical != self.p_is_radical


__all__ = ['DynamicElement']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from ...exceptions import EmptyMolecule


common_isotopes = {'H': 1, 'He': 4, 'Li': 7, 'Be': 9, 'B': 11, 'C': 12, 'N': 14, 'O': 16, 'F': 19, 'Ne': 20, 'Na': 23,
                   'Mg': 24, 'Al': 27, 'Si': 28, 'P': 31, 'S': 32, 'Cl': 35, 'Ar': 40, 'K': 39, 'Ca': 40, 'Sc': 45,
                   'Ti': 48, 'V': 51, 'Cr': 52, 'Mn': 55, 'Fe': 56, 'Co': 59, 'Ni': 59, 'Cu': 64, 'Zn': 65, 'Ga': 70,
                   'Ge': 73, 'As': 75, 'Se': 79, 'Br': 80, 'Kr': 84, 'Rb': 85, 'Sr': 88, 'Y': 89, 'Zr': 91, 'Nb': 93,
                   'Mo': 96, 'Tc': 98, 'Ru': 101, 'Rh': 103, 'Pd': 106, 'Ag': 108, 'Cd': 112, 'In': 115, 'Sn': 119,
                   'Sb': 122, 'Te': 128, 'I': 127, 'Xe': 131, 'Cs': 133, 'Ba': 137, 'La': 139, 'Ce': 140, 'Pr': 141,
                   'Nd': 144, 'Pm': 145, 'Sm': 150, 'Eu': 152, 'Gd': 157, 'Tb': 159, 'Dy': 163, 'Ho': 165, 'Er': 167,
                   'Tm': 169, 'Yb': 173, 'Lu': 175, 'Hf': 178, 'Ta': 181, 'W': 184, 'Re': 186, 'Os': 190, 'Ir': 192,
                   'Pt': 195, 'Au': 197, 'Hg': 201, 'Tl': 204, 'Pb': 207, 'Bi': 209, 'Po': 209, 'At': 210, 'Rn': 222,
                   'Fr': 223, 'Ra': 226, 'Ac': 227, 'Th': 232, 'Pa': 231, 'U': 238, 'Np': 237, 'Pu': 244, 'Am': 243,
                   'Cm': 247, 'Bk': 247, 'Cf': 251, 'Es': 252, 'Fm': 257, 'Md': 258, 'No': 259, 'Lr': 260, 'Rf': 261,
                   'Db': 270, 'Sg': 269, 'Bh': 270, 'Hs': 270, 'Mt': 278, 'Ds': 281, 'Rg': 281, 'Cn': 285, 'Nh': 278,
                   'Fl': 289, 'Mc': 289, 'Lv': 293, 'Ts': 297, 'Og': 294}
_ctf_data = {'R': 'is_radical', 'C': 'charge', 'I': 'isotope'}
_charge_map = {'  0': 0, '  1': 3, '  2': 2, '  3': 1, '  4': 0, '  5': -1, '  6': -2, '  7': -3}


class MOLRead:
    def __init__(self, line, log_buffer=None):
        self.__atoms_count = int(line[0:3])
        self.__bonds_count = int(line[3:6])
        self.__dat = {}
        self.__atoms = []
        self.__bonds = []
        self.__stereo = []
        self.__hydrogens = {}
        if log_buffer is None:
            log_buffer = []
        self.__log_buffer = log_buffer

    def __new__(cls, line, log_buffer=None):
        if line.startswith('  0'):
            raise EmptyMolecule
        return super().__new__(cls)

    def getvalue(self):
        if self.__mend:
            return {'atoms': self.__atoms, 'bonds': self.__bonds, 'stereo': self.__stereo, 'meta': {},
                    'hydrogens': self.__hydrogens}
        raise ValueError('molecule not complete')

    def __call__(self, line):
        if self.__mend:
            raise ValueError('parser closed')
        elif len(self.__atoms) < self.__atoms_count:
            try:
                charge = _charge_map[line[36:39]]
            except KeyError:
                raise ValueError('invalid charge')
            element = line[31:34].strip()
            isotope = line[34:36]

            if element in 'AL':
                raise ValueError('queries not supported')
            elif element == 'D':
                element = 'H'
                if isotope != ' 0':
                    raise ValueError('isotope on deuterium atom')
                isotope = 2
            elif isotope != ' 0':
                try:
                    isotope = common_isotopes[element] + int(isotope)
                except KeyError:
                    raise ValueError('invalid element symbol')
            else:
                isotope = None

            mapping = line[60:63]
            self.__atoms.append({'element': element, 'charge': charge, 'isotope': isotope, 'is_radical': False,
                                 'mapping': int(mapping) if mapping else 0,
                                 'x': float(line[0:10]), 'y': float(line[10:20]), 'z': float(line[20:30])})

        elif len(self.__bonds) < self.__bonds_count:
            a1, a2 = int(line[0:3]) - 1, int(line[3:6]) - 1
            s = line[9:12]
            if s == '  1':
                self.__stereo.append((a1, a2, 1))
            elif s == '  6':
                self.__stereo.append((a1, a2, -1))
            elif s != '  0':
                self.__log_buffer.append('unsupported or invalid stereo')
            b = int(line[6:9])
            if b == 9:  # added ad-hoc for bond type 9
                b = 8
                self.__log_buffer.append('coordinate bond replaced with special')
            self.__bonds.append((a1, a2, b))
        elif line.startswith('M  END'):
            for a in self.__atoms:
                if a['is_radical']:
                    a['is_radical'] = True
            for x in self.__dat.values():
                try:
                    _type = x['type']
                    if _type == 'mrv_implicit_h':
                        atoms = x['atoms']
                        value = x['value']
                        if len(atoms) != 1 or atoms[0] == -1 or not value:
                            raise ValueError(f'MRV_IMPLICIT_H spec invalid {x}')
                        self.__hydrogens[atoms[0]] = int(value[6:])
                    else:
                        self.__log_buffer.append(f'ignored data: {x}')
                except KeyError:
                    raise ValueError(f'Invalid SGROUP {x}')
            self.__mend = True
            return True
        else:
            self.__collect(line)

    def __collect(self, line):
        if line.startswith('M  ALS'):
            raise ValueError('list of atoms not supported')
        elif line.startswith(('M  ISO', 'M  RAD', 'M  CHG')):
            _type = _ctf_data[line[3]]
            for i in range(int(line[6:9])):
                i8 = i * 8
                atom = int(line[10 + i8:13 + i8])
                if not atom or atom > len(self.__atoms):
                    raise ValueError('invalid atoms number')
                atom = self.__atoms[atom - 1]
                atom[_type] = int(line[14 + i8:17 + i8])

        elif line.startswith('M  STY'):
            for i in range(int(line[6:9])):
                i8 = i * 8
                if 'DAT' == line[14 + i8:17 + i8]:
                    self.__dat[int(line[10 + i8:13 + i8])] = {}
        elif line.startswith('M  SAL'):
            i = int(line[7:10])
            if i in self.__dat:
                self.__dat[i]['atoms'] = tuple(int(line[14 + 4 * i:17 + 4 * i]) - 1 for i in range(int(line[10:13])))
        elif line.startswith('M  SDT'):
            i = int(line[7:10])
            if i in self.__dat:
                self.__dat[i]['type'] = line.split()[-1].lower()
        elif line.startswith('M  SED'):
            i = int(line[7:10])
            if i in self.__dat:
                self.__dat[i]['value'] = line[10:].strip().replace('/', '').lower()

    __mend = False


__all__ = ['MOLRead', 'common_isotopes']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from chython import smiles
from itertools import zip_longest


def test_keto_enol_2h_pyrrole():
    """
    2H‐pyrrole. [N:1]=1[C:2][C:3]=,:[C:4][C:5]=1
    """
    for t, v in zip(['C1C=CC=N1', 'C1N=CC2=C1C=CC=C2'], ['C=1C=CNC=1', 'N1C=C2C=CC=CC2=C1']):
        s = smiles(t)
        t = set(s.enumerate_tautomers())
        v = smiles(v)
        s.thiele()
        v.thiele()
        assert len(t) == 2, ' '.join(str(x) for x in t)
        assert t == {s, v}, f'{", ".join(str(x) for x in t)} != {s}, {v}'


def test_acid_protonated_nitrogen():
    """
    Ammonia, H-imine,guanidine,amidine. [H][N+]=,-,:
    """
    for t, v in zip_longest(['C=[NH2+].[Cl-]', 'C=[NH+]C.[Cl-]', 'C[NH3+].[Cl-]', 'C[NH2+]C.[Cl-]', 'C[NH1+](C)C.[Cl-]',
                             '[NH4+].[Cl-]', 'C=[N+](C)C.[Cl-]', 'C[N+](C)(C)C.[Cl-]'],
                            ['C=N.Cl', 'C=NC.Cl', 'CN.Cl', 'CNC.Cl', 'CN(C)C.Cl', 'N.Cl']):
        s = smiles(t)
        t = set(s.enumerate_tautomers())
        if v:
            assert len(t) == 2, ' '.join(str(x) for x in t)
            assert t == {s, smiles(v)}
        else:
            assert len(t) == 1


def test_base_nitrogen():
    for t, v in zip_longest(['N1C=CN=N1.Cl', 'NC(N)=N.Cl', 'CN(C)C(=NO)N(C)C.Cl', 'CN(C)C(=NC)N(C)C.Cl',
                             'CN(C)C(=NN)N(C)C.Cl', 'COC(N)=N.Cl', 'CSC(N)=N.Cl', 'COC(OC)=N.Cl', 'COC(C)=N.Cl',
                             'CNN.Cl', 'CN.Cl',
                             'N.Cl', 'N1C=CC=C1.Cl'],
                            [('N1N=NC=C1.Cl', 'N1=CC=NN1.Cl', '[NH+]=1NC=CN=1.[Cl-]', 'N1=[NH+]C=CN1.[Cl-]',
                              'N1=CC=[NH+]N1.[Cl-]'),
                             ('NC(N)=N.Cl', 'NC(N)=[NH2+].[Cl-]'),
                             ('CN(C)C(=NO)N(C)C.Cl', 'CN(C)C(N(C)C)=[NH+]O.[Cl-]'),
                             ('CN(C)C(=NC)N(C)C.Cl', 'CN(C)C(N(C)C)=[NH+]C.[Cl-]'),
                             ('CN(C)C(=NN)N(C)C.Cl', 'CN(C)C(N(C)C)=[NH+]N.[Cl-]'),
                             ('Cl.NC(=N)OC', '[NH2+]=C(N)OC.[Cl-]'), ('Cl.NC(=N)SC', '[NH2+]=C(N)SC.[Cl-]'),
                             ('COC(OC)=N.Cl', 'COC(OC)=[NH2+].[Cl-]'),
                             ('COC(C)=N.Cl', 'COC(C)=[NH2+].[Cl-]', 'C(N)(OC)=C.Cl'),
                             ('CNN.Cl', 'CN[NH3+].[Cl-]', 'C[NH2+]N.[Cl-]'),
                             ('CN.Cl', 'C[NH3+].[Cl-]')]):
        s = smiles(t)
        t = set(s.enumerate_tautomers(full=True))
        if v:
            assert len(t) == len(v), ' '.join(str(x) for x in t)
            vs = set()
            for x in v:
                x = smiles(x)
                x.thiele(fix_tautomers=False)
                vs.add(x)
            assert t == vs, ' '.join(str(x) for x in t) + ' != ' + ' '.join(str(x) for x in vs)
        else:
            assert len(t) == 1, ' '.join(str(x) for x in t)



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_method
from functools import reduce
from hashlib import sha512
from itertools import chain
from operator import or_
from typing import Dict, Iterable, Iterator, Optional, Tuple
from .cgr import CGRContainer
from .molecule import MoleculeContainer
from ..algorithms.calculate2d import Calculate2DReaction
from ..algorithms.depict import DepictReaction
from ..algorithms.standardize import StandardizeReaction


class ReactionContainer(StandardizeReaction, Calculate2DReaction, DepictReaction):
    """
    Reaction storage. Contains reactants, products and reagents lists.

    Reaction storage hashable and comparable. based on reaction unique signature (SMILES).
    """
    __slots__ = ('__reactants', '__products', '__reagents', '__meta', '__name', '_arrow', '_signs', '__dict__')
    __class_cache__ = {}

    def __init__(self, reactants: Iterable[MoleculeContainer] = (), products: Iterable[MoleculeContainer] = (),
                 reagents: Iterable[MoleculeContainer] = (), meta: Optional[Dict] = None, name: Optional[str] = None):
        """
        New reaction object creation

        :param reactants: list of MoleculeContainers in left side of reaction
        :param products: right side of reaction. see reactants
        :param reagents: middle side of reaction: solvents, catalysts, etc. see reactants
        :param meta: dictionary of metadata. like DTYPE-DATUM in RDF

        """
        reactants = tuple(reactants)
        products = tuple(products)
        reagents = tuple(reagents)
        if not reactants and not products and not reagents:
            raise ValueError('At least one graph object required')
        elif not all(isinstance(x, MoleculeContainer) for x in chain(reactants, products, reagents)):
            raise TypeError(f'MoleculeContainers expected')

        self.__reactants = reactants
        self.__products = products
        self.__reagents = reagents
        if meta is None:
            self.__meta = None
        else:
            self.__meta = dict(meta)
        if name is None:
            self.__name = None
        else:
            self.name = name
        self._arrow = None
        self._signs = None

    @property
    def reactants(self) -> Tuple[MoleculeContainer, ...]:
        return self.__reactants

    @property
    def reagents(self) -> Tuple[MoleculeContainer, ...]:
        return self.__reagents

    @property
    def products(self) -> Tuple[MoleculeContainer, ...]:
        return self.__products

    def molecules(self) -> Iterator[MoleculeContainer]:
        """
        Iterator of all reaction molecules
        """
        return chain(self.__reactants, self.__reagents, self.__products)

    @property
    def meta(self) -> Dict:
        """
        Dictionary of metadata.
        Like DTYPE-DATUM in RDF
        """
        if self.__meta is None:
            self.__meta = {}  # lazy
        return self.__meta

    @property
    def name(self) -> str:
        return self.__name or ''

    @name.setter
    def name(self, name: str):
        if not isinstance(name, str):
            raise TypeError('name should be string up to 80 symbols')
        self.__name = name

    def copy(self) -> 'ReactionContainer':
        """
        Get copy of object
        """
        copy = object.__new__(self.__class__)
        copy._ReactionContainer__reactants = tuple(x.copy() for x in self.__reactants)
        copy._ReactionContainer__products = tuple(x.copy() for x in self.__products)
        copy._ReactionContainer__reagents = tuple(x.copy() for x in self.__reagents)
        copy._ReactionContainer__meta = self.__meta.copy()
        copy._ReactionContainer__name = self.__name
        copy._arrow = self._arrow
        copy._signs = self._signs
        return copy

    @cached_method
    def compose(self) -> CGRContainer:
        """
        Get CGR of reaction

        Reagents will be presented as unchanged molecules
        :return: CGRContainer
        """
        rr = self.__reagents + self.__reactants
        if rr:
            r = reduce(or_, rr)
        else:
            r = MoleculeContainer()
        if self.__products:
            p = reduce(or_, self.__products)
        else:
            p = MoleculeContainer()
        return r ^ p

    @classmethod
    def from_cgr(cls, cgr: CGRContainer) -> 'ReactionContainer':
        """
        Decompose CGR into reaction
        """
        if not isinstance(cgr, CGRContainer):
            raise TypeError('CGR expected')
        r, p = ~cgr
        return ReactionContainer(r, p)

    def flush_cache(self):
        self.__dict__.clear()
        for m in self.molecules():
            m.flush_cache()

    def __invert__(self) -> CGRContainer:
        """
        Get CGR of reaction
        """
        return self.compose()

    def __eq__(self, other):
        return isinstance(other, ReactionContainer) and str(self) == str(other)

    @cached_method
    def __hash__(self):
        return hash(str(self))

    @cached_method
    def __bytes__(self):
        return sha512(str(self).encode()).digest()

    def __bool__(self):
        """
        Exists both reactants and products
        """
        return bool(self.__reactants and self.__products)

    @cached_method
    def __str__(self):
        return format(self)

    def __format__(self, format_spec):
        """
        :param format_spec: see specification of nested containers.
            !c - Keep nested containers order
            !C - skip cxsmiles fragments contract
        """
        sig = []
        count = 0
        contract = []
        for ml in (self.__reactants, self.__reagents, self.__products):
            if not format_spec or '!c' not in format_spec:
                ml = sorted(ml, key=str)
            for m in ml:
                if m.connected_components_count > 1:
                    contract.append([str(x + count) for x in range(m.connected_components_count)])
                    count += m.connected_components_count
                else:
                    count += 1
            if format_spec:
                sig.append('.'.join(format(x, format_spec) for x in ml))
            else:
                sig.append('.'.join(str(x) for x in ml))
        if (not format_spec or '!C' not in format_spec) and contract:
            return f"{'>'.join(sig)} |f:{','.join('.'.join(x) for x in contract)}|"
        return '>'.join(sig)

    @cached_method
    def __len__(self):
        return len(self.__reactants) + len(self.__products) + len(self.__reagents)

    def __getstate__(self):
        state = {'reactants': self.__reactants, 'products': self.__products, 'reagents': self.__reagents,
                 'meta': self.__meta, 'name': self.__name, 'arrow': self._arrow, 'signs': self._signs}
        import chython
        if chython.pickle_cache:
            state['cache'] = self.__dict__
        return state

    def __setstate__(self, state):
        self.__reactants = state['reactants']
        self.__products = state['products']
        self.__reagents = state['reagents']
        self.__meta = state['meta']
        self.__name = state['name']
        self._arrow = state['arrow']
        self._signs = state['signs']
        if 'cache' in state:
            self.__dict__.update(state['cache'])


__all__ = ['ReactionContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy


def _rules():
    from ...containers import QueryContainer
    raw_rules = []

    # Nitro
    #
    #      O
    #     //
    # * - N+
    #      \
    #       O-
    #
    atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 2, 'charge': 1},
             {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Carbonate
    #
    #      O
    #     //
    # * - C
    #      \
    #       O-
    #
    atoms = ({'atom': 'C', 'neighbors': 3, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1, 'charge': -1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Carbon Acid
    #
    #      O
    #     //
    # * - C
    #      \
    #       OH
    #
    atoms = ({'atom': 'C', 'neighbors': 3, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Phosphonic acid
    #
    #      *
    #      |
    #  * - P = O
    #      |
    #      OH
    #
    atoms = ({'atom': 'P', 'neighbors': 4, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Phosphate
    #
    #      *
    #      |
    #  * - P = O
    #      |
    #      O-
    #
    atoms = ({'atom': 'P', 'neighbors': 4, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1, 'charge': -1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Sulphonic acid
    #
    #       O
    #      //
    #  * - S = O
    #      |
    #      OH
    #
    atoms = ({'atom': 'S', 'neighbors': 4}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Sulphonate
    #
    #       O
    #      //
    #  * - S = O
    #      |
    #      O-
    #
    atoms = ({'atom': 'S', 'neighbors': 4}, {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (atoms, bonds), fix))

    # Nitro addition
    #
    #      O             O -- *
    #     //            /
    # * - N+   >>  * = N+
    #      \            \
    #       O-           O-
    #
    atoms = ({'atom': 'N', 'neighbors': 3, 'charge': 1, 'hybridization': 2},
             {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    p_atoms = ({'atom': 'N', 'neighbors': 3, 'charge': 1, 'hybridization': 2},
               {'atom': 'O', 'neighbors': 1, 'charge': -1}, {'atom': 'O', 'neighbors': 2})
    p_bonds = ((1, 2, 1), (1, 3, 1))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (p_atoms, p_bonds), fix))

    # Sulphate addition
    #
    #      O [3]            O -- * [2]
    #     //               /
    # * = S - *   >>  * = S - *
    #     |               \\
    #     O- [2]           O [3]
    #
    atoms = ({'atom': 'S', 'neighbors': 4, 'hybridization': 3}, {'atom': 'O', 'neighbors': 1, 'charge': -1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    p_atoms = ({'atom': 'S', 'neighbors': 4, 'hybridization': 3}, {'atom': 'O', 'neighbors': 2},
               {'atom': 'O', 'neighbors': 1})
    p_bonds = ((1, 2, 1), (1, 3, 2))
    fix = {2: 3, 3: 2}
    raw_rules.append(((atoms, bonds), (p_atoms, p_bonds), fix))

    compiled_rules = []
    for (r_atoms, r_bonds), (p_atoms, p_bonds), fix in raw_rules:
        r_q = QueryContainer()
        p_q = QueryContainer()
        for a in r_atoms:
            r_q.add_atom(**a)
        for n, m, b in r_bonds:
            r_q.add_bond(n, m, b)
        for a in p_atoms:
            p_q.add_atom(**a)
        for n, m, b in p_bonds:
            p_q.add_bond(n, m, b)
        compiled_rules.append((r_q, p_q, fix))
    return compiled_rules


rules = Proxy(_rules)


__all__ = ['rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2020 Dinar Batyrshin <batyrshin-dinar@mail.ru>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#


def functional_groups(molecule, limit):
    """
    Generate all connected atom groups up to limit atoms.

    :param molecule: MoleculeContainer
    :param limit: chain length
    :return: list of molecule functional groups
    """
    bonds = molecule._bonds

    if limit < 1:
        raise ValueError('limit should be >= 1')

    response = []
    groups = set()
    stack = [([a], list(n)) for a, n in bonds.items()]
    while stack:
        aug, nei = stack.pop(0)
        for x in nei:
            augx = (*aug, x)
            if augx not in groups:
                groups.add(augx)
                response.append(molecule.substructure(augx, as_query=True))
                nt = nei.copy()
                nt.remove(x)
                nt.extend(list(bonds[x]))
                if len(augx) < limit:
                    stack.append((augx, nt))
    return response


__all__ = ['functional_groups']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import zip_longest
from typing import List, Tuple, Union
from ..containers import MoleculeContainer, QueryContainer, CGRContainer

graphs = Union[MoleculeContainer, QueryContainer, CGRContainer]


def grid_depict(molecules: Union[List[graphs], Tuple[graphs, ...]], cols: int = 3):
    """
    Depict molecules grid.

    :param cols: number of molecules per row.
    """
    config = MoleculeContainer._render_config
    font_size = config['font_size']
    font125 = 1.25 * font_size

    planes = []
    render = []
    shift_y = 2.
    shift_x = 0.
    for ms in zip_longest(*[iter(molecules)] * cols):
        height = 0.
        for m in ms:
            if m is None:
                break
            min_y = min(y for x, y in m._plane.values())
            max_y = max(y for x, y in m._plane.values())
            h = max_y - min_y
            if height < h:  # get height of row
                height = h
            planes.append(m._plane.copy())

        # now align mols by height
        shift_y -= height / 2. + 2.
        max_x = 0.
        for m in ms:
            if m is None:
                break
            max_x = m._fix_plane_mean(max_x, shift_y) + 2.
            render.append(m.depict(embedding=True)[:5])
            if max_x > shift_x:  # get total width
                shift_x = max_x
        shift_y -= height / 2.

    # restore planes
    for p, m in zip(planes, molecules):
        m._plane = p

    width = shift_x + 3.0 * font_size
    height = -shift_y + 2.5 * font_size
    svg = [f'<svg width="{width:.2f}cm" height="{height:.2f}cm" '
           f'viewBox="{-font125:.2f} {-font125:.2f} {width:.2f} '
           f'{height:.2f}" xmlns="http://www.w3.org/2000/svg" version="1.1">']
    for atoms, bonds, define, masks, uid in render:
        svg.extend(MoleculeContainer._graph_svg(atoms, bonds, define, masks, uid, -font125, -font125, width, height))
    svg.append('</svg>')
    return '\n'.join(svg)


__all__ = ['grid_depict']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from functools import cached_property
from typing import Dict, Iterator, List, Tuple
from . import molecule  # cyclic imports resolve
from .bonds import DynamicBond
from ..algorithms.calculate2d import Calculate2DCGR
from ..algorithms.depict import DepictCGR
from ..algorithms.isomorphism import Isomorphism
from ..algorithms.morgan import Morgan
from ..algorithms.rings import Rings
from ..algorithms.smiles import CGRSmiles
from ..algorithms.x3dom import X3domCGR
from ..periodictable import DynamicElement, Element


class CGRContainer(CGRSmiles, Morgan, Rings, Isomorphism, DepictCGR, Calculate2DCGR, X3domCGR):
    __slots__ = ('_atoms', '_bonds', '_charges', '_radicals', '_p_charges', '_p_radicals', '_plane', '_conformers',
                 '__dict__', '__weakref__')
    _atoms: Dict[int, DynamicElement]
    _bonds: Dict[int, Dict[int, DynamicBond]]
    _charges: Dict[int, int]
    _radicals: Dict[int, bool]
    _p_charges: Dict[int, int]
    _p_radicals: Dict[int, bool]
    _plane: Dict[int, Tuple[float, float]]
    _conformers: List[Dict[int, Tuple[float, float, float]]]

    def __init__(self):
        self._atoms = {}
        self._bonds = {}
        self._charges = {}
        self._radicals = {}
        self._p_charges = {}
        self._p_radicals = {}
        self._plane = {}
        self._conformers = []

    def bonds(self) -> Iterator[Tuple[int, int, DynamicBond]]:
        """
        Iterate other all bonds
        """
        seen = set()
        for n, m_bond in self._bonds.items():
            seen.add(n)
            for m, bond in m_bond.items():
                if m not in seen:
                    yield n, m, bond

    @cached_property
    def center_atoms(self) -> Tuple[int, ...]:
        """ Get list of atoms of reaction center (atoms with dynamic: bonds, charges, radicals).
        """
        radicals = self._radicals
        p_charges = self._p_charges
        p_radicals = self._p_radicals

        center = set()
        for n, c in self._charges.items():
            if c != p_charges[n] or radicals[n] != p_radicals[n]:
                center.add(n)

        for n, m_bond in self._bonds.items():
            if any(bond.order != bond.p_order for bond in m_bond.values()):
                center.add(n)

        return tuple(center)

    @cached_property
    def aromatic_rings(self) -> Tuple[Tuple[int, ...], ...]:
        """
        existed or formed aromatic rings atoms numbers
        """
        adj = self._bonds
        return tuple(ring for ring in self.sssr if
                     adj[ring[0]][ring[-1]].order == 4 and all(adj[n][m].order == 4 for n, m in zip(ring, ring[1:])) or
                     adj[ring[0]][ring[-1]].p_order == 4 and all(adj[n][m].p_order == 4 for n, m in zip(ring, ring[1:]))
                     )

    def substructure(self, atoms) -> 'CGRContainer':
        """
        Create substructure containing atoms from atoms list

        :param atoms: list of atoms numbers of substructure
        """
        atoms = set(atoms)
        sa = self._atoms
        sc = self._charges
        sr = self._radicals
        sp = self._plane
        sb = self._bonds
        spc = self._p_charges
        spr = self._p_radicals

        sub = object.__new__(self.__class__)
        sub._charges = {n: sc[n] for n in atoms}
        sub._radicals = {n: sr[n] for n in atoms}
        sub._p_charges = {n: spc[n] for n in atoms}
        sub._p_radicals = {n: spr[n] for n in atoms}
        sub._plane = {n: sp[n] for n in atoms}
        sub._conformers = [{n: c[n] for n in atoms} for c in self._conformers]

        sub._atoms = ca = {}
        for n in atoms:
            ca[n] = atom = sa[n].copy()
            atom._attach_to_graph(sub, n)

        sub._bonds = cb = {}
        for n in atoms:
            cb[n] = cbn = {}
            for m, bond in sb[n].items():
                if m in cb:  # bond partially exists. need back-connection.
                    cbn[m] = cb[m][n]
                elif m in atoms:
                    cbn[m] = bond.copy()
        return sub

    def decompose(self) -> Tuple['molecule.MoleculeContainer', 'molecule.MoleculeContainer']:
        """
        decompose CGR to pair of Molecules, which represents reactants and products state of reaction

        :return: tuple of two molecules
        """
        charges = self._charges
        p_charges = self._p_charges
        radicals = self._radicals
        p_radicals = self._p_radicals
        plane = self._plane

        reactants = molecule.MoleculeContainer()
        products = molecule.MoleculeContainer()

        for n, atom in self._atoms.items():
            atom = Element.from_atomic_number(atom.atomic_number)(atom.isotope)
            reactants.add_atom(atom, n, charge=charges[n], is_radical=radicals[n], xy=plane[n])
            products.add_atom(atom.copy(), n, charge=p_charges[n], is_radical=p_radicals[n], xy=plane[n])

        for n, m, bond in self.bonds():
            if bond.order:
                reactants.add_bond(n, m, bond.order)
            if bond.p_order:
                products.add_bond(n, m, bond.p_order)
        return reactants, products

    def get_mapping(self, other: 'CGRContainer', /, **kwargs):
        if isinstance(other, CGRContainer):
            return super().get_mapping(other, **kwargs)
        raise TypeError('CGRContainer expected')

    def __iter__(self):
        return iter(self._atoms)

    def __invert__(self):
        """
        decompose CGR
        """
        return self.decompose()

    def __getstate__(self):
        return {'atoms': self._atoms, 'bonds': self._bonds, 'plane': self._plane, 'conformers': self._conformers,
                'charges': self._charges, 'radicals': self._radicals,
                'p_charges': self._p_charges, 'p_radicals': self._p_radicals}

    def __setstate__(self, state):
        self._atoms = state['atoms']
        for n, a in state['atoms'].items():
            a._attach_to_graph(self, n)
        self._charges = state['charges']
        self._radicals = state['radicals']
        self._plane = state['plane']
        self._bonds = state['bonds']
        self._p_charges = state['p_charges']
        self._p_radicals = state['p_radicals']
        self._conformers = state['conformers']


__all__ = ['CGRContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from ._write import _MDLWrite
from ...containers import MoleculeContainer


class EMDLWrite(_MDLWrite):
    def _convert_molecule(self, g, write3d=None):
        if not isinstance(g, MoleculeContainer):
            raise TypeError('MoleculeContainer expected')

        head = f'M  V30 BEGIN CTAB\nM  V30 COUNTS {g.atoms_count} {g.bonds_count} 0 0 0'

        bond = ['M  V30 BEGIN BOND']
        mapping = {m: n for n, m in enumerate(g, start=1)}
        wedge = defaultdict(set)
        bonds = g._bonds
        i = 0
        for i, (n, m, s) in enumerate(g._wedge_map, start=1):
            bond.append(f'M  V30 {i} {bonds[n][m].order} {mapping[n]} {mapping[m]} CFG={s == 1 and "1" or "3"}')
            wedge[n].add(m)
            wedge[m].add(n)

        for i, (n, m, b) in enumerate(g.bonds(), start=i + 1):
            if m not in wedge[n]:
                bond.append(f'M  V30 {i} {b.order} {mapping[n]} {mapping[m]}')
        bond.append('M  V30 END BOND\nM  V30 END CTAB\n')
        bond = '\n'.join(bond)

        if write3d is not None:
            return '\n'.join((head, self.__convert_atoms3d(g, g._conformers[write3d]), bond))
        else:
            return '\n'.join((head, self.__convert_atoms2d(g), bond))

    def __convert_atoms2d(self, g):
        gc = g._charges
        gr = g._radicals
        gp = g._plane

        out = ['M  V30 BEGIN ATOM']
        for n, (m, a) in enumerate(g._atoms.items(), start=1):
            x, y = gp[m]
            c = gc[m]
            c = f' CHG={c}' if c else ''
            r = ' RAD=2' if gr[m] else ''
            i = f' MASS={a.isotope}' if a.isotope else ''

            if self._mapping:
                out.append(f'M  V30 {n} {a.atomic_symbol} {x:.4f} {y:.4f} 0 {m}{c}{r}{i}')
            else:
                out.append(f'M  V30 {n} {a.atomic_symbol} {x:.4f} {y:.4f} 0 0{c}{r}{i}')
        out.append('M  V30 END ATOM')
        return '\n'.join(out)

    def __convert_atoms3d(self, g, xyz):
        gc = g._charges
        gr = g._radicals

        out = ['M  V30 BEGIN ATOM']
        for n, (m, a) in enumerate(g._atoms.items(), start=1):
            x, y, z = xyz[m]
            c = gc[m]
            c = f' CHG={c}' if c else ''
            r = ' RAD=2' if gr[m] else ''
            i = f' MASS={a.isotope}' if a.isotope else ''

            if self._mapping:
                out.append(f'M  V30 {n} {a.atomic_symbol} {x:.4f} {y:.4f} {z:.4f} {m}{c}{r}{i}')
            else:
                out.append(f'M  V30 {n} {a.atomic_symbol} {x:.4f} {y:.4f} {z:.4f} 0{c}{r}{i}')
        out.append('M  V30 END ATOM')
        return '\n'.join(out)


__all__ = ['EMDLWrite']
